#include "sprox_desfire_i.h"

/*
 * DESFire cryptographic functions
 * -------------------------------
 */

static DES_TDES_CTX_ST local_context;

/* DES part */
/* -------- */

/* Compute a DES MAC in CBC receive mode */
static void DesFire_DES_MAC(DES_TDES_CTX_ST *context, BYTE *data, WORD blocks, const BYTE * key, BYTE * result)
{
  WORD    i;
  BYTE    j;
  BYTE    carry[8], buffer[8];
  
  if (context == NULL) context = &local_context;

  DES_Init(&context->des, (BYTE *) key);
  memset(carry, 0x00, 8);       /* IV <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    memcpy(buffer, &data[8 * i], 8);
    for (j = 0; j < 8; j++)
      buffer[j] ^= carry[j];    /* P  <- P XOR IV  */
    DES_Encrypt2(&context->des, carry, buffer); /* IV <- DES(P)    */
  }
  memcpy(result, carry, 4);     /* This is the MAC */
}

/* DES decryption in CBC send mode */
static void DesFire_DES_Send(DES_TDES_CTX_ST *context, BYTE *data, WORD blocks, const BYTE * key)
{
  WORD    i;
  BYTE    j;
  BYTE    carry[8];
  
  if (context == NULL) context = &local_context;

  DES_Init(&context->des, (BYTE *) key);
  memset(carry, 0x00, 8);       /* IV <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    for (j = 0; j < 8; j++)
      data[8 * i + j] ^= carry[j];  /* P  <- P XOR IV  */
    DES_Decrypt2(&context->des, carry, &data[8 * i]); /* IV <- iDES(P)   */
    memcpy(&data[8 * i], carry, 8); /* P  <- IV        */
  }
}

/* DES decryption in CBC receive mode */
static void DesFire_DES_Recv(DES_TDES_CTX_ST *context, BYTE *data, WORD blocks, const BYTE * key)
{
  WORD    i;
  BYTE    j;
  BYTE    carry[8], buffer[8];
  
  if (context == NULL) context = &local_context;

  DES_Init(&context->des, (BYTE *) key);
  memset(buffer, 0x00, 8);      /* B  <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    memcpy(carry, buffer, 8);   /* IV <- B         */
    memcpy(buffer, &data[8 * i], 8);  /* B  <- P         */
    DES_Decrypt2(&context->des, &data[8 * i], buffer);  /* P  <- iDES(B)   */
    for (j = 0; j < 8; j++)
      data[8 * i + j] ^= carry[j];  /* P  <- P XOR IV  */
  }
}

/* 3DES part */
/* --------- */

/* Compute a 3DES MAC in CBC receive mode */
static void DesFire_TDES_MAC(DES_TDES_CTX_ST *context, BYTE *data, WORD blocks, const BYTE * key, BYTE * result)
{
  WORD    i;
  BYTE    j;
  BYTE    carry[8], buffer[8];

  if (context == NULL) context = &local_context;

  TDES_Init(&context->tdes, (BYTE *) & key[0], (BYTE *) & key[8], NULL);
  memset(carry, 0x00, 8);       /* IV <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    memcpy(buffer, &data[8 * i], 8);
    for (j = 0; j < 8; j++)
      buffer[j] ^= carry[j];    /* P  <- P XOR IV  */
    TDES_Encrypt2(&context->tdes, carry, buffer);  /* IV <- 3DES(P)   */
  }
  memcpy(result, carry, 4);     /* This is the MAC */
}



/* 3DES decryption in CBC send mode */
static void DesFire_TDES_Send(DES_TDES_CTX_ST *context, BYTE *data, WORD blocks, const BYTE * key)
{
  DWORD   i;
  BYTE    j;
  BYTE    carry[8];
  
  if (context == NULL) context = &local_context;

  TDES_Init(&context->tdes, (BYTE *) & key[0], (BYTE *) & key[8], NULL);
  memset(carry, 0x00, 8);       /* IV <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    for (j = 0; j < 8; j++)
      data[8 * i + j] ^= carry[j];  /* P  <- P XOR IV  */
    TDES_Decrypt2(&context->tdes, carry, &data[8 * i]);  /* IV <- i3DES(P)   */
    memcpy(&data[8 * i], carry, 8); /* P  <- IV        */
  }
}

/* 3DES decryption in CBC receive mode */
static void DesFire_TDES_Recv(DES_TDES_CTX_ST *context, BYTE *data, WORD blocks, const BYTE * key)
{
  DWORD   i;
  BYTE    j;
  BYTE    carry[8], buffer[8];
  
  if (context == NULL) context = &local_context;

  TDES_Init(&context->tdes, (BYTE *) & key[0], (BYTE *) & key[8], NULL);
  memset(buffer, 0x00, 8);      /* B  <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    memcpy(carry, buffer, 8);   /* IV <- B         */
    memcpy(buffer, &data[8 * i], 8);  /* B  <- P         */
    TDES_Decrypt2(&context->tdes, &data[8 * i], buffer); /* P  <- iDES(B)   */
    for (j = 0; j < 8; j++)
      data[8 * i + j] ^= carry[j];  /* P  <- P XOR IV  */
  }
}


/*
 * Compute the MAC for DESFire MACED communication
 * -----------------------------------------------
 */
SPROX_API_FUNC_T(void, Desfire_CipherMAC) (SPROX_PARAM  BYTE *data, WORD blocks, const BYTE * key, BYTE * result)
{
  BYTE    mac32[4];
  SPROX_DESFIRE_GET_CTX_V();

  /* Well formatted buffer */
  if (memcmp(&key[0], &key[8], 8))
  {
    /* 3DES */
    DesFire_TDES_MAC(&ctx->cipher_context, data, blocks, key, mac32);
  } else
  {
    /* DES */
    DesFire_DES_MAC(&ctx->cipher_context, data, blocks, key, mac32);
  }

  if (result != NULL)
    memcpy(result, mac32, 4);
}

SPROX_API_FUNC_T(void, Desfire_CipherMACb) (SPROX_PARAM  BYTE *data, DWORD size, const BYTE * key, BYTE * result)
{
  SPROX_DESFIRE_GET_CTX_V();
  
  if (size % 8)
  {
    WORD    blocks;

    memset(ctx->cipher_buffer, 0, sizeof(ctx->cipher_buffer));
    memcpy(ctx->cipher_buffer, data, size);

    blocks = 1 + (WORD) (size / 8);

    SPROX_API_CALL(Desfire_CipherMAC) (SPROX_PARAM_P  ctx->cipher_buffer, blocks, key, result);
  } else
  {
    SPROX_API_CALL(Desfire_CipherMAC) (SPROX_PARAM_P  data, (WORD) (size / 8), key, result);
  }
}

/*
 * Decipher data received from the DESFire
 * ---------------------------------------
 */
SPROX_API_FUNC_T(void, Desfire_CipherRecv) (SPROX_PARAM  BYTE *data, WORD blocks, const BYTE * key)
{
  SPROX_DESFIRE_GET_CTX_V();
  
  if (memcmp(&key[0], &key[8], 8))
  {
    /* 3DES */
    DesFire_TDES_Recv(&ctx->cipher_context, data, blocks, key);
  } else
  {
    /* DES */
    DesFire_DES_Recv(&ctx->cipher_context, data, blocks, key);
  }
}

/*
 * Encipher data to be sent to the DESFire
 * ---------------------------------------
 */
SPROX_API_FUNC_T(void, Desfire_CipherSend) (SPROX_PARAM  BYTE *data, WORD blocks, const BYTE * key)
{
  SPROX_DESFIRE_GET_CTX_V();
  
  if (memcmp(&key[0], &key[8], 8))
  {
    /* 3DES */
    DesFire_TDES_Send(&ctx->cipher_context, data, blocks, key);
  } else
  {
    /* DES */
    DesFire_DES_Send(&ctx->cipher_context, data, blocks, key);
  }
}
