#include "sprox_desfire_i.h"
#ifndef UNDER_CE
  #include <time.h>
#endif

/*
 * Authenticate
 * ------------
 */
SPROX_API_FUNC(Desfire_Authenticate) (SPROX_PARAM  BYTE bKeyNumber, const BYTE pbAccessKey[16])
{
  BOOL    qIsTripleDes = FALSE;
  SPROX_RC   status, w;
//  BYTE abCbcBlock[8];
  BYTE    abRndB[8], abRndA[8];
  SPROX_DESFIRE_GET_CTX();

  /* Each new Authenticate must invalidate the current authentication status. */
  ctx->auth_key = INVALIDATE_AUTHENTICATION;

  /* Check whether a TripleDES key was passed (both key halves are different). */
  if (memcmp(pbAccessKey, pbAccessKey + 8, 8))
  {
    /* If the two key halves are not identical, we are doing TripleDES. */

    /* We have to remember that TripleDES is in effect, because the manner of building
       the session key is different in this case. */
    qIsTripleDes = TRUE;
  }

  /* Create the command string consisting of the command byte and the parameter byte. */
  ctx->buffer[INF + 0] = DF_AUTHENTICATE;
  ctx->buffer[INF + 1] = bKeyNumber;
  ctx->buflen = 2;

  /* Send the command string to the PICC and get its response (1st frame exchange).
     The PICC has to respond with an DF_ADDITIONAL_FRAME status byte. */
  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_ADDITIONAL_FRAME);
  if (status != DF_OPERATION_OK)
    return status;

  /* Check the number of bytes received, we expect 9 bytes. */
  if (ctx->buflen != 9)
  {
    /* Error: block with inappropriate number of bytes received from the card. */
    return DFCARD_WRONG_LENGTH;
  }

  /* OK, we received the 8 bytes Ek( RndB ) from the PICC.
     Decipher Ek( RndB ) to get RndB in ctx->buffer.
     Note that the status code has already been extracted from the queue. */
  SPROX_API_CALL(Desfire_CipherRecv) (SPROX_PARAM_P  &ctx->buffer[INF + 1], 1, pbAccessKey);
  /* Store this RndB (is needed later on for generating the session key). */
  memcpy(abRndB, &ctx->buffer[INF + 1], 8);

  /* Now the PCD has to generate RndA.
     But first seed the random number generator with the current time. */
#ifndef UNDER_CE
  srand((unsigned) time(NULL));
#else
  /* No time on Windows CE */
  srand(0xdead);
#endif
  for (w = 0; w < 8; w++)
    abRndA[w] = (BYTE) rand();

#if 0
  for (w = 0; w < 8; w += 2)
  {
    abRndA[w] = 0x45;
    abRndA[w + 1] = 0x67;
  }
#endif

  D(printf("RndA: "));
  D(dump_buffer(abRndA, 8));
  D(printf("\n"));
  D(printf("RndB: "));
  D(dump_buffer(abRndB, 8));
  D(printf("\n"));

  /* Start the second frame with a status byte indicating to the PICC that the Authenticate
     command is continued. */
  ctx->buffer[INF + 0] = DF_ADDITIONAL_FRAME;

  /* Append RndA and RndB' ( RndB' is generated by rotating RndB one byte to the left )
     after the status byte. */
  memcpy(&ctx->buffer[INF + 1], abRndA, 8);
  memcpy(&ctx->buffer[INF + 9], &abRndB[1], 7);
  ctx->buffer[INF + 16] = abRndB[0]; /* first byte move to last byte */
  ctx->buflen = 17;

  /* Apply the DES send operation to the 16 argument bytes before sending the second frame
     to the PICC ( do not include the status byte in the DES operation ). */
  SPROX_API_CALL(Desfire_CipherSend) (SPROX_PARAM_P  &ctx->buffer[INF + 1], 2, pbAccessKey);

  /* Send the 2nd frame to the PICC and get its response. */
  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;

  /* We should now have Ek( RndA' ) in our buffer.
     RndA' was made from RndA by the PICC by rotating the string one byte to the left.
     Decipher Ek( RndA' ) to get RndA' in ctx->buffer. */
  SPROX_API_CALL(Desfire_CipherRecv) (SPROX_PARAM_P  &ctx->buffer[INF + 1], 2, pbAccessKey);

  /* Now we have RndA' in our buffer.
     We have to check whether it matches our local copy of RndA.
     If one of the subsequent comparisons fails, we do not trust the PICC and therefore
     abort the authentication procedure ( no session key is generated ). */

  /* First compare the bytes 1 to 7 of RndA with the first 7 bytes in the queue. */
  if (memcmp(&ctx->buffer[INF + 1], abRndA + 1, 7))
    return DFCARD_WRONG_KEY;

  /* Then compare the leftmost byte of RndA with the last byte in the queue. */
  if (ctx->buffer[INF + 8] != abRndA[0])
    return DFCARD_WRONG_KEY;

  /* The actual authentication has succeeded.
     Finally we have to generate the session key from both random numbers RndA and RndB.
     The first half of the session key is the concatenation of RndA[0-3] + RndB[0-3]. */
  memcpy(ctx->session_key, abRndA, 4);
  memcpy(ctx->session_key + 4, abRndB, 4);

  /* If the original key passed through pbAccessKey is a TripleDES key, the session
     key must also be a TripleDES key. */
  if (qIsTripleDes)
  {
    /* For TripleDES generate the second part of the session key.
       This is the concatenation of RndA[4-7] + RndB[4-7]. */
    memcpy(ctx->session_key + 8, abRndA + 4, 4);
    memcpy(ctx->session_key + 12, abRndB + 4, 4);
  } else
  {
    memcpy(ctx->session_key + 8, abRndA, 4);
    memcpy(ctx->session_key + 12, abRndB, 4);
  }

  D(printf("Session Key: "));
  D(dump_buffer(ctx->session_key, 16));
  D(printf("\n"));

  /* Authenticate succeeded, therefore we remember the number of the key which was used
     to obtain the current authentication status. */
  ctx->auth_key = bKeyNumber;

  /* Success. */
  return DF_OPERATION_OK;
}

/*
 * ChangeKeySettings
 * -----------------
 */
SPROX_API_FUNC(Desfire_ChangeKeySettings) (SPROX_PARAM  BYTE key_settings)
{
  BYTE    i;
  SPROX_DESFIRE_GET_CTX();

  /* Create the info block containing the command code */
  ctx->buffer[INF + 0] = DF_CHANGE_KEY_SETTINGS;
  ctx->buffer[INF + 1] = key_settings;
  ctx->buflen = 9;               /* 1 byte cmd, 8 bytes of decipher setting */

  /* Append the CRC value corresponding to the key_settings byte */
  /* CRC16 calculation according to 14443A                       */
  ComputeIso14443ACrc(&ctx->buffer[INF + 2], &ctx->buffer[INF + 1], 1);

  D(printf("Change Key Settings 1: "));
  D(dump_buffer(&ctx->buffer[INF + 1], 3));
  D(printf("\n"));

  /* 'Encrypt' the bNewKeySettings byte and its CRC bytes padding the DES block with 0 bytes. */
  /* Note: since mifare DESFire always runs in DES ENcryption mode this is actually a DES     */
  /* DEcryption operation.                                                                    */
  for (i = 4; i < 9; i++)
    ctx->buffer[INF + i] = 0;
  SPROX_API_CALL(Desfire_CipherSend) (SPROX_PARAM_P  &ctx->buffer[INF + 1], 1, ctx->session_key);

  D(printf("Change Key Settings 2: "));
  D(dump_buffer(&ctx->buffer[INF + 2], 8));
  D(printf("\n"));

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * GetKeySettings
 * --------------
 */
SPROX_API_FUNC(Desfire_GetKeySettings) (SPROX_PARAM  BYTE *key_settings, BYTE *key_count)
{
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* Create the info block containing the command code */
  ctx->buffer[INF + 0] = DF_GET_KEY_SETTINGS;
  ctx->buflen = 1;

  /* Communicate the info block to the card and check the operation's return status. */
  /* GetKeySettings returns 3 bytes, the status and two bytes of information.        */
  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 3, ACCEPT_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;

  /* Return the requested key settings bytes. */
  *key_settings = ctx->buffer[INF + 1];
  *key_count = ctx->buffer[INF + 2];

  /* Success */
  return DF_OPERATION_OK;
}

/*
 * ChangeKey
 * ---------
 */
SPROX_API_FUNC(Desfire_ChangeKey) (SPROX_PARAM  BYTE key_number, const BYTE new_key[16], const BYTE old_key[16])
{
  BYTE    b, buffer[24];
  SPROX_DESFIRE_GET_CTX();

  /* Begin the info block with the command code and the number of the key to be changed. */
  ctx->buffer[INF + 0] = DF_CHANGE_KEY;
  ctx->buffer[INF + 1] = key_number;
  ctx->buflen = 26;

  D(printf("Change Key, new key = "));
  D(dump_buffer(new_key, 16));
  D(printf("\n"));

  if (old_key != NULL)
  {
    /* If a 'previous key' has been passed, format the 24 byte cryptogram according to the     */
    /* three key procedure (new key, previous key, all encrypted with the current session key) */

    /* Take the new key into the buffer. */
    memcpy(buffer, new_key, 16);

    /* XOR the previous key to the new key. */
    for (b = 0; b < 16; b++)
      buffer[b] ^= old_key[b];

    /* Append a first CRC, computed over the XORed key combination. */
    ComputeIso14443ACrc(&buffer[16], &buffer[0], 16);

    /* Append a second CRC, computed over the new key only. */
    ComputeIso14443ACrc(&buffer[18], (BYTE *) new_key, 16);

    /* Fill the remainder of the 3rd block with zeros (4 bytes). */
    memset(&buffer[20], 0x00, 4);
  } else
  {
    /* If no 'previous key' has been passed, format the 24 byte cryptogram according to the */
    /* two key procedure (new key, encrypted with the current session key).                 */

    /* Take the new key into the buffer. */
    memcpy(buffer, new_key, 16);

    /* Append the CRC computed over the new key. */
    ComputeIso14443ACrc(&buffer[16], &buffer[0], 16);

    /* fill the remainder of the 3rd block with zeros (6 bytes). */
    memset(&buffer[18], 0x00, 6);
  }

  /* Append the 24 byte buffer to the command string. */
  memcpy(&ctx->buffer[INF + 2], buffer, 24);

  D(printf("Change Key 1: "));
  D(dump_buffer(&ctx->buffer[INF + 2], 24));
  D(printf("\n"));

  /* 'Encrypt' the 24 key, CRC and padding bytes in CBC mode.                             */
  /* Note: since mifare DESFire always runs in DES ENcryption mode this is actually a DES */
  /* DEcryption operation.                                                                */
  SPROX_API_CALL(Desfire_CipherSend) (SPROX_PARAM_P  &ctx->buffer[INF + 2], 3, ctx->session_key);

  D(printf("Change Key 2: "));
  D(dump_buffer(&ctx->buffer[INF + 2], 24));
  D(printf("\n"));

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * GetKeyVersion
 * -------------
 */
SPROX_API_FUNC(Desfire_GetKeyVersion) (SPROX_PARAM  BYTE bKeyNumber, BYTE *pbKeyVersion)
{
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* Create the info block containing the command code and the key number argument. */
  ctx->buffer[INF + 0] = DF_GET_KEY_VERSION;
  ctx->buffer[INF + 1] = bKeyNumber;
  ctx->buflen = 2;

  /* Communicate the info block to the card and check the operation's return status.   */
  /* The returned info block must contain two bytes, the status code and the requested */
  /* key version byte.                                                                 */
  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 2, ACCEPT_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;

  /* Get the key version byte. */
  *pbKeyVersion = ctx->buffer[INF + 1];

  /* Success. */
  return DF_OPERATION_OK;
}

/*
 * FormatPICC
 * ----------
 */
SPROX_API_FUNC(Desfire_FormatPICC) (SPROX_PARAM_V)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code. */
  ctx->buffer[INF + 0] = DF_FORMAT_PICC;
  ctx->buflen = 1;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * CreateApplication
 * -----------------
 */
SPROX_API_FUNC(Desfire_CreateApplication) (SPROX_PARAM  DWORD aid, BYTE key_settings, BYTE keys_count)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code and the given parameters. */
  ctx->buffer[INF + 0] = DF_CREATE_APPLICATION;
  ctx->buffer[INF + 1] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->buffer[INF + 2] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->buffer[INF + 3] = (BYTE) (aid & 0x000000FF);
  ctx->buffer[INF + 4] = key_settings;
  ctx->buffer[INF + 5] = keys_count;
  ctx->buflen = 6;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * DeleteApplication
 * -----------------
 */
SPROX_API_FUNC(Desfire_DeleteApplication) (SPROX_PARAM  DWORD aid)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code and the given parameters. */
  ctx->buffer[INF + 0] = DF_DELETE_APPLICATION;
  ctx->buffer[INF + 1] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->buffer[INF + 2] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->buffer[INF + 3] = (BYTE) (aid & 0x000000FF);
  ctx->buflen = 4;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * GetApplicationIDs
 * -----------------
 */
SPROX_API_FUNC(Desfire_GetApplicationIDs) (SPROX_PARAM  BYTE aid_max_count, DWORD aid_list[], BYTE *aid_count)
{
  DWORD   aid;
  BYTE    i;
  WORD    len;
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* Set the number of applications to zero */
  *aid_count = 0;

  /* create the info block containing the command code */
  ctx->buffer[INF + 0] = DF_GET_APPLICATION_IDS;
  ctx->buflen = 1;

  for (;;)
  {
    status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_ADDITIONAL_FRAME | ACCEPT_OPERATION_OK);
    if (status != DF_OPERATION_OK)
      return status;

    len = ctx->buflen - 1;       /* substract 1 byte for the received status */

    /* ByteCount must be in multiples of APPLICATION_ID_SIZE bytes
       we check this to proof the format integrity
       if zero bytes have been received this is ok -> means no
       applications existing */
    if (len % APPLICATION_ID_SIZE)
      return DFCARD_WRONG_LENGTH;

    for (i = 0; i < (len / APPLICATION_ID_SIZE); i++)
    {
      /* Extract AID */
      aid = ctx->buffer[INF + 3 + APPLICATION_ID_SIZE * i];
      aid <<= 8;
      aid += ctx->buffer[INF + 2 + APPLICATION_ID_SIZE * i];
      aid <<= 8;
      aid += ctx->buffer[INF + 1 + APPLICATION_ID_SIZE * i];

      /* Copy AID */
      aid_list[*aid_count] = aid;

      /* Next AID */
      *aid_count = *aid_count + 1;
      if (*aid_count >= aid_max_count)
      {
        if (i < ((len / APPLICATION_ID_SIZE) - 1))
        {
          /* Still ApplicationIDs available but we ran out
             of buffer space provided by the caller */
          return DFCARD_OVERFLOW;
        }
      }
    }

    /* Prepare the info block indicating that we want more application IDs */
    if (ctx->buffer[INF + 0] != DF_ADDITIONAL_FRAME)
      return DF_OPERATION_OK;
    ctx->buflen = 1;
  }
}

/*
 * SelectApplication
 * -----------------
 */
SPROX_API_FUNC(Desfire_SelectApplication) (SPROX_PARAM  DWORD aid)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Each SelectApplication causes a currently valid authentication state to be lost */
  ctx->auth_key = INVALIDATE_AUTHENTICATION;

  /* Create the info block containing the command code and the given parameters. */
  ctx->buffer[INF + 0] = DF_SELECT_APPLICATION;
  ctx->buffer[INF + 1] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->buffer[INF + 2] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->buffer[INF + 3] = (BYTE) (aid & 0x000000FF);
  ctx->buflen = 4;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * GetVersion
 * ----------
 */
SPROX_API_FUNC(Desfire_GetVersion) (SPROX_PARAM  DF_VERSION_INFO *pVersionInfo)
{
  SPROX_RC   status;
  BYTE   *ptr = (BYTE *) pVersionInfo;
  SPROX_DESFIRE_GET_CTX();

  ctx->buffer[INF + 0] = DF_GET_VERSION;
  ctx->buflen = 1;
  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 8, ACCEPT_ADDITIONAL_FRAME);
  if (status != DF_OPERATION_OK)
    return status;
  memcpy(ptr, &ctx->buffer[INF + 1], 7);
  ptr += 7;

  ctx->buffer[INF + 0] = DF_ADDITIONAL_FRAME;
  ctx->buflen = 1;
  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 8, ACCEPT_ADDITIONAL_FRAME);
  if (status != DF_OPERATION_OK)
    return status;
  memcpy(ptr, &ctx->buffer[INF + 1], 7);
  ptr += 7;

  ctx->buffer[INF + 0] = DF_ADDITIONAL_FRAME;
  ctx->buflen = 1;
  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 15, ACCEPT_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;
  memcpy(ptr, &ctx->buffer[INF + 1], 14);

  return DF_OPERATION_OK;
}

/*
 * GetFileIDs
 * ----------
 */
SPROX_API_FUNC(Desfire_GetFileIDs) (SPROX_PARAM  BYTE fid_max_count, BYTE fid_list[], BYTE *fid_count)
{
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* reset the number of found FileIDs */
  *fid_count = 0;

  /* create the info block containing the command code */
  ctx->buffer[INF + 0] = DF_GET_FILE_IDS;
  ctx->buflen = 1;

  /* this command terminates within a single frame (no additional frames necessary) */
  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;

  ctx->buflen--;                 /* subtract 1 byte for the received status of desfire card */
  for (; (*fid_count < fid_max_count) && (ctx->buflen > 0); fid_list++, (*fid_count)++)
  {
    *fid_list = ctx->buffer[INF + 1 + *fid_count];
    ctx->buflen--;
  }

  /* if there are still bytes left in the ctx->buffer   */
  /* we ran out of buffer space provided by the caller */
  if (ctx->buflen != 0)
    return DFCARD_OVERFLOW;

  return DF_OPERATION_OK;
}

/*
 * GetFileSettings
 * ---------------
 */
SPROX_API_FUNC(Desfire_GetFileSettings) (SPROX_PARAM  BYTE file_id, BYTE *file_type, BYTE *comm_mode, WORD *access_rights, DF_ADDITIONAL_FILE_SETTINGS *additionnal_settings)
{
  BYTE    l_file_type;
  SPROX_RC   status;
  DWORD   temp;
  SPROX_DESFIRE_GET_CTX();

  if (file_type == NULL)
    file_type = &l_file_type;

  /* create command code and append FileID parameter */
  ctx->buffer[INF + 0] = DF_GET_FILE_SETTINGS;
  ctx->buffer[INF + 1] = file_id;
  ctx->buflen = 2;

  /* don't do any implicit length check on the response */
  /* this has to be resolved depending on the file type */
  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;

  /* there is a fixed section of the response which is common to all file types */
  /* therefore the response must at least be of this fixed length               */
  /* Note the status byte had to be extract!                                    */

  ctx->buflen--;                 /* subtract 1 byte for the received status of desfire card */
  if (ctx->buflen < DF_FILE_SETTINGS_COMMON_LENGTH)
    return DFCARD_WRONG_LENGTH;

  /* extract the file's type */
  *file_type = ctx->buffer[INF + 1];

  /* extract the file's communication mode */
  if (comm_mode != NULL)
    *comm_mode = ctx->buffer[INF + 2];

  /* extract the file's access rights */
  if (access_rights != NULL)
  {
    *access_rights = ctx->buffer[INF + 4];
    *access_rights <<= 8;
    *access_rights += ctx->buffer[INF + 3];
  }

  /* if no pointer to a union for holding the file settings has been passed */
  /* we empty the buffer and return with success                            */
  if (additionnal_settings == NULL)
    return DF_OPERATION_OK;

  /* the rest of the buffer contains file type specific information      */
  /* therefore we have to resolve it according to the received file type */
  ctx->buflen -= DF_FILE_SETTINGS_COMMON_LENGTH; /* subtract for the fixed section */
  switch (*file_type)
  {
    case DF_STANDARD_DATA_FILE:
    case DF_BACKUP_DATA_FILE:

      if (ctx->buflen != DF_DATA_FILE_SETTINGS_LENGTH)
        return DFCARD_WRONG_LENGTH;

      /* extract the file size */
      temp = ctx->buffer[INF + 7];
      temp <<= 8;
      temp += ctx->buffer[INF + 6];
      temp <<= 8;
      temp += ctx->buffer[INF + 5];
      additionnal_settings->stDataFileSettings.eFileSize = temp;
      break;

    case DF_VALUE_FILE:

      if (ctx->buflen != DF_VALUE_FILE_SETTINGS_LENGTH)
        return DFCARD_WRONG_LENGTH;

      /* extract the lower limit */
      temp = ctx->buffer[INF + 8];
      temp <<= 8;
      temp += ctx->buffer[INF + 7];
      temp <<= 8;
      temp += ctx->buffer[INF + 6];
      temp <<= 8;
      temp += ctx->buffer[INF + 5];
      additionnal_settings->stValueFileSettings.lLowerLimit = temp;

      /* extract the upper limit */
      temp = ctx->buffer[INF + 12];
      temp <<= 8;
      temp += ctx->buffer[INF + 11];
      temp <<= 8;
      temp += ctx->buffer[INF + 10];
      temp <<= 8;
      temp += ctx->buffer[INF + 9];
      additionnal_settings->stValueFileSettings.lUpperLimit = temp;

      /* extract the limited credit value; */
      temp = ctx->buffer[INF + 16];
      temp <<= 8;
      temp += ctx->buffer[INF + 15];
      temp <<= 8;
      temp += ctx->buffer[INF + 14];
      temp <<= 8;
      temp += ctx->buffer[INF + 13];
      additionnal_settings->stValueFileSettings.eLimitedCredit = temp;

      /* limited Credit enabled ? */
      additionnal_settings->stValueFileSettings.bLimitedCreditEnabled = ctx->buffer[INF + 17];
      break;

    case DF_LINEAR_RECORD_FILE:
    case DF_CYCLIC_RECORD_FILE:

      if (ctx->buflen != DF_RECORD_FILE_SETTINGS_LENGTH)
        return DFCARD_WRONG_LENGTH;

      /* extract the number of entire records */
      temp = ctx->buffer[INF + 7];
      temp <<= 8;
      temp += ctx->buffer[INF + 6];
      temp <<= 8;
      temp += ctx->buffer[INF + 5];
      additionnal_settings->stRecordFileSettings.eRecordSize = temp;

      /* get the size of one record */
      temp = ctx->buffer[INF + 10];
      temp <<= 8;
      temp += ctx->buffer[INF + 9];
      temp <<= 8;
      temp += ctx->buffer[INF + 8];
      additionnal_settings->stRecordFileSettings.eMaxNRecords = temp;

      /* get the current number of records */
      temp = ctx->buffer[INF + 13];
      temp <<= 8;
      temp += ctx->buffer[INF + 12];
      temp <<= 8;
      temp += ctx->buffer[INF + 11];
      additionnal_settings->stRecordFileSettings.eCurrNRecords = temp;

      break;

    default:
      /* Unknown file type */
      return DFCARD_WRONG_FILE_TYPE;
  }

  return DF_OPERATION_OK;
}

/* 
 * ChangeFileSettings
 * ------------------
 */
SPROX_API_FUNC(Desfire_ChangeFileSettings) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, WORD access_rights)
{
  BOOL    comm_encrypted;
  SPROX_RC   status;
  WORD    old_access_rights;
  SPROX_DESFIRE_GET_CTX();

  /* Before being able to send the new file settings to the PICC, we need to know whether */
  /* the current access rights setting requires the communication to be encrypted.        */
  status = SPROX_API_CALL(Desfire_GetFileSettings) (SPROX_PARAM_P  file_id, NULL, NULL, &old_access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;
  comm_encrypted = (old_access_rights & DF_CHANGE_ACCESS_RIGHTS);

  /* Create the info block containing the command code and the given parameters. */
  ctx->buffer[INF + 0] = DF_CHANGE_FILE_SETTINGS;
  ctx->buffer[INF + 1] = file_id;
  ctx->buffer[INF + 2] = comm_mode;
  ctx->buffer[INF + 3] = (BYTE) (access_rights & 0x00FF);
  access_rights >>= 8;
  ctx->buffer[INF + 4] = (BYTE) (access_rights & 0x00FF);
  ctx->buflen = 5;

  if (comm_encrypted != 0x0E)
  {
    /* If encrypted communication is requested, append a CRC, computed over the communication mode         */
    /* and access rights arguments (3bytes), to the command string. Then 'Encrypt' these arguments and the */
    /* CRC to ensure that only an authenticated PCD can change the settings.                               */
    /* Note: since mifare DESFire always runs in DES ENcryption mode this is actually a DES                */
    /* DEcryption operation.                                                                               */
    ComputeIso14443ACrc(&ctx->buffer[INF + 5], &ctx->buffer[INF + 2], 3);
    memset(&ctx->buffer[INF + 7], 0x00, 3);
    SPROX_API_CALL(Desfire_CipherSend) (SPROX_PARAM_P  &ctx->buffer[INF + 2], 1, ctx->session_key);
    ctx->buflen += 5;
  }

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * CreateStdDataFile
 * -----------------
 */
SPROX_API_FUNC(Desfire_CreateStdDataFile) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, WORD access_rights, DWORD file_size)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code and the given parameters. */
  ctx->buffer[INF + 0] = DF_CREATE_STD_DATA_FILE;
  ctx->buffer[INF + 1] = file_id;
  ctx->buffer[INF + 2] = comm_mode;
  ctx->buffer[INF + 3] = (BYTE) (access_rights & 0x00FF);
  access_rights >>= 8;
  ctx->buffer[INF + 4] = (BYTE) (access_rights & 0x00FF);
  ctx->buffer[INF + 5] = (BYTE) (file_size & 0x00FF);
  file_size >>= 8;
  ctx->buffer[INF + 6] = (BYTE) (file_size & 0x00FF);
  file_size >>= 8;
  ctx->buffer[INF + 7] = (BYTE) (file_size & 0x00FF);
  ctx->buflen = 8;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * CreateBackupDataFile
 * --------------------
 */
SPROX_API_FUNC(Desfire_CreateBackupDataFile) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, WORD access_rights, DWORD file_size)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code and the given parameters. */
  ctx->buffer[INF + 0] = DF_CREATE_BACKUP_DATA_FILE;
  ctx->buffer[INF + 1] = file_id;
  ctx->buffer[INF + 2] = comm_mode;
  ctx->buffer[INF + 3] = (BYTE) (access_rights & 0x00FF);
  access_rights >>= 8;
  ctx->buffer[INF + 4] = (BYTE) (access_rights & 0x00FF);
  ctx->buffer[INF + 5] = (BYTE) (file_size & 0x000000FF);
  file_size >>= 8;
  ctx->buffer[INF + 6] = (BYTE) (file_size & 0x000000FF);
  file_size >>= 8;
  ctx->buffer[INF + 7] = (BYTE) (file_size & 0x000000FF);
  ctx->buflen = 8;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * CreateValueFile
 * ---------------
 */
SPROX_API_FUNC(Desfire_CreateValueFile) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, WORD access_rights, SDWORD lower_limit, SDWORD upper_limit, SDWORD initial_value, BYTE limited_credit_enabled)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code and the given parameters. */
  ctx->buffer[INF + 0] = DF_CREATE_VALUE_FILE;
  ctx->buffer[INF + 1] = file_id;
  ctx->buffer[INF + 2] = comm_mode;
  ctx->buffer[INF + 3] = (BYTE) (access_rights & 0x00FF);
  access_rights >>= 8;
  ctx->buffer[INF + 4] = (BYTE) (access_rights & 0x00FF);
  ctx->buffer[INF + 5] = (BYTE) (lower_limit & 0x000000FF);
  lower_limit >>= 8;
  ctx->buffer[INF + 6] = (BYTE) (lower_limit & 0x000000FF);
  lower_limit >>= 8;
  ctx->buffer[INF + 7] = (BYTE) (lower_limit & 0x000000FF);
  lower_limit >>= 8;
  ctx->buffer[INF + 8] = (BYTE) (lower_limit & 0x000000FF);
  ctx->buffer[INF + 9] = (BYTE) (upper_limit & 0x000000FF);
  upper_limit >>= 8;
  ctx->buffer[INF + 10] = (BYTE) (upper_limit & 0x000000FF);
  upper_limit >>= 8;
  ctx->buffer[INF + 11] = (BYTE) (upper_limit & 0x000000FF);
  upper_limit >>= 8;
  ctx->buffer[INF + 12] = (BYTE) (upper_limit & 0x000000FF);
  ctx->buffer[INF + 13] = (BYTE) (initial_value & 0x000000FF);
  initial_value >>= 8;
  ctx->buffer[INF + 14] = (BYTE) (initial_value & 0x000000FF);
  initial_value >>= 8;
  ctx->buffer[INF + 15] = (BYTE) (initial_value & 0x000000FF);
  initial_value >>= 8;
  ctx->buffer[INF + 16] = (BYTE) (initial_value & 0x000000FF);
  ctx->buffer[INF + 17] = limited_credit_enabled;
  ctx->buflen = 18;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * CreateLinearRecordFile
 * ----------------------
 */
SPROX_API_FUNC(Desfire_CreateLinearRecordFile) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, WORD access_rights, DWORD record_size, DWORD max_records)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code and the given parameters. */
  ctx->buffer[INF + 0] = DF_CREATE_LINEAR_RECORD_FILE;
  ctx->buffer[INF + 1] = file_id;
  ctx->buffer[INF + 2] = comm_mode;
  ctx->buffer[INF + 3] = (BYTE) (access_rights & 0x00FF);
  access_rights >>= 8;
  ctx->buffer[INF + 4] = (BYTE) (access_rights & 0x00FF);
  ctx->buffer[INF + 5] = (BYTE) (record_size & 0x000000FF);
  record_size >>= 8;
  ctx->buffer[INF + 6] = (BYTE) (record_size & 0x000000FF);
  record_size >>= 8;
  ctx->buffer[INF + 7] = (BYTE) (record_size & 0x000000FF);
  ctx->buffer[INF + 8] = (BYTE) (max_records & 0x000000FF);
  max_records >>= 8;
  ctx->buffer[INF + 9] = (BYTE) (max_records & 0x000000FF);
  max_records >>= 8;
  ctx->buffer[INF + 10] = (BYTE) (max_records & 0x000000FF);
  ctx->buflen = 11;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * CreateCyclicRecordFile
 * ----------------------
 */

SPROX_API_FUNC(Desfire_CreateCyclicRecordFile) (SPROX_PARAM  BYTE file_id, BYTE comm_mode, WORD access_rights, DWORD record_size, DWORD max_records)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code and the given parameters. */
  ctx->buffer[INF + 0] = DF_CREATE_CYCLIC_RECORD_FILE;
  ctx->buffer[INF + 1] = file_id;
  ctx->buffer[INF + 2] = comm_mode;
  ctx->buffer[INF + 3] = (BYTE) (access_rights & 0x00FF);
  access_rights >>= 8;
  ctx->buffer[INF + 4] = (BYTE) (access_rights & 0x00FF);
  ctx->buffer[INF + 5] = (BYTE) (record_size & 0x000000FF);
  record_size >>= 8;
  ctx->buffer[INF + 6] = (BYTE) (record_size & 0x000000FF);
  record_size >>= 8;
  ctx->buffer[INF + 7] = (BYTE) (record_size & 0x000000FF);
  ctx->buffer[INF + 8] = (BYTE) (max_records & 0x000000FF);
  max_records >>= 8;
  ctx->buffer[INF + 9] = (BYTE) (max_records & 0x000000FF);
  max_records >>= 8;
  ctx->buffer[INF + 10] = (BYTE) (max_records & 0x000000FF);
  ctx->buflen = 11;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * DeleteFile
 * ----------
 */
SPROX_API_FUNC(Desfire_DeleteFile) (SPROX_PARAM  BYTE file_id)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code and the given parameters. */
  ctx->buffer[INF + 0] = DF_DELETE_FILE;
  ctx->buffer[INF + 1] = file_id;
  ctx->buflen = 2;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}


/*
 * CommitTransaction
 * -----------------
 */
SPROX_API_FUNC(Desfire_CommitTransaction) (SPROX_PARAM_V)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code. */
  ctx->buffer[INF + 0] = DF_COMMIT_TRANSACTION;
  ctx->buflen = 1;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * AbortTransaction
 * ----------------
 */
SPROX_API_FUNC(Desfire_AbortTransaction) (SPROX_PARAM_V)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code. */
  ctx->buffer[INF + 0] = DF_ABORT_TRANSACTION;
  ctx->buflen = 1;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}
