#include "cardappz.h"
#include "products/springprox/springprox.h"

#include <assert.h>
#include <stdio.h>

#define CHECK_NONE 0
#define CHECK_OK   1
#define CHECK_MORE 2

/**f* Desfire/TDES_Decrypt_CBC_S
 *
 * NAME
 *   TDES_Decrypt_CBC_S
 *
 * DESCRIPTION
 *   Triple-DES decryption, in CBC SEND mode
 *   This is specific to Desfire cards
 *
 * SYNOPSIS
 *   void TDES_Decrypt_CBC_S(TDES_CTX_ST *tdes_ctx, unsigned char *data, int blocks);
 *
 * INPUTS
 *   TDES_CTX_ST   *tdes_ctx           : a previoulsy inited 3-DES context
 *   unsigned char *data               : the data to decrypt (in place processing)
 *   int           blocks              : the number of 8-byte blocks of data
 *
 * RETURNS
 *   None
 *
 **/
void TDES_Decrypt_CBC_S(TDES_CTX_ST * tdes_ctx, unsigned char *data,
                        int blocks)
{
  int i, j;
  unsigned char carry[8];

  assert(tdes_ctx != NULL);
  assert(data != NULL);

  memset(carry, 0x00, 8);       /* IV <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    for (j = 0; j < 8; j++)
      data[8 * i + j] ^= carry[j];  /* P  <- P XOR IV  */
    TDES_Decrypt(tdes_ctx, &data[8 * i]); /* P <- i3DES(P)   */
    memcpy(carry, &data[8 * i], 8); /* P  <- IV        */
  }
}

/**f* Desfire/TDES_Encrypt_CBC_S
 *
 * NAME
 *   TDES_Encrypt_CBC_S
 *
 * DESCRIPTION
 *   Triple-DES encryption, in CBC SEND mode
 *   This is standard 3-DES CBC implementation
 *
 * SYNOPSIS
 *   void TDES_Encrypt_CBC_S(TDES_CTX_ST *tdes_ctx, unsigned char *data, int blocks);
 *
 * INPUTS
 *   TDES_CTX_ST   *tdes_ctx           : a previoulsy inited 3-DES context
 *   unsigned char *data               : the data to encrypt (in place processing)
 *   int           blocks              : the number of 8-byte blocks of data
 *
 * RETURNS
 *   None
 *
 **/
void TDES_Encrypt_CBC_S(TDES_CTX_ST * tdes_ctx, unsigned char *data,
                        int blocks)
{
  int i, j;
  unsigned char carry[8];

  assert(tdes_ctx != NULL);
  assert(data != NULL);

  memset(carry, 0x00, 8);       /* IV <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    for (j = 0; j < 8; j++)
      data[8 * i + j] ^= carry[j];  /* P  <- P XOR IV  */
    TDES_Encrypt(tdes_ctx, &data[8 * i]); /* P <- i3DES(P)   */
    memcpy(carry, &data[8 * i], 8); /* P  <- IV        */
  }
}

/**f* Desfire/TDES_Decrypt_CBC_R
 *
 * NAME
 *   TDES_Decrypt_CBC_R
 *
 * DESCRIPTION
 *   Triple-DES decryption, in CBC RECEIVE mode
 *   This is standard 3-DES CBC implementation, also used with Desfire cards
 *
 * INPUTS
 *   TDES_CTX_ST   *tdes_ctx           : a previoulsy inited 3-DES context
 *   unsigned char *data               : the data to decrypt (in place processing)
 *   int           blocks              : the number of 8-byte blocks of data
 *
 * RETURNS
 *   None
 *
 **/
void TDES_Decrypt_CBC_R(TDES_CTX_ST * tdes_ctx, unsigned char *data,
                        int blocks)
{
  int i, j;
  unsigned char carry[8], buffer[8];

  assert(tdes_ctx != NULL);
  assert(data != NULL);

  memset(buffer, 0x00, 8);      /* B  <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    memcpy(carry, buffer, 8);   /* IV <- B         */
    memcpy(buffer, &data[8 * i], 8);  /* B  <- P         */
    TDES_Decrypt(tdes_ctx, &data[8 * i]); /* P  <- iDES(B)   */
    for (j = 0; j < 8; j++)
      data[8 * i + j] ^= carry[j];  /* P  <- P XOR IV  */
  }
}

/**f* Desfire/DesfireDecipherDataEx
 *
 * NAME
 *   DesfireDecipherDataEx
 *
 * DESCRIPTION
 *   Desfire Triple-DES receive operation
 *   Decipher frame, check padding and CRC
 *
 * SYNOPSIS
 *   int DesfireDecipherDataEx(TDES_CTX_ST   *tdes_ctx,
 *                    [in/out] unsigned char *data,
 *                             int           recv_length,
 *                    [output] int           *real_length,
 *                             int           padd80);
 *
 * NOTES
 *   int recv_length  : size of incoming data buffer
 *   int *real_length : [output] actual size of data, after removal of padding and CRC
 *   int padd80       : set to 1 if data must be padded with 0x80 ... 0x00
 *                      set to 0 if data must be padded with 0x00 ... 0x00
 *
 **/
int DesfireDecipherDataEx(TDES_CTX_ST * tdes_ctx, unsigned char *data,
                          int recv_length, int *real_length, int padd80)
{
  int c = 0;
  unsigned char crc[2];

  assert(tdes_ctx != NULL);
  assert(data != NULL);
  assert(real_length != NULL);

  /* Length must be a multiple of 8 (block size) */
  if ((recv_length < 8) || (recv_length % 8))
    return CARDAPPZ_CARD_SECURITY_ERROR;

  /* Decipher in place */
  TDES_Decrypt_CBC_R(tdes_ctx, data, recv_length / 8);

  /* Check padding, either 00 .. 00 or 80 ... 00 if asked by padd80 */
  while (data[recv_length - 1] == 0x00)
  {
    recv_length--;
    c++;
  }
  if (padd80)
  {
    if (data[recv_length - 1] != 0x80)
      return CARDAPPZ_CARD_SECURITY_ERROR;  /* Wrong padding */
    recv_length--;
    c++;
  }

  /* Padding must be between 1 and 8 */
  if ((c < 1) || (c > 8))
    return CARDAPPZ_CARD_SECURITY_ERROR;  /* Wrong padding */

  /* Buffer is ended with 2 CRC bytes */
  recv_length -= 2;
  SPROX_ComputeIso14443ACrc(crc, data, (WORD) recv_length);
  if (memcmp(crc, &data[recv_length], 2))
    return CARDAPPZ_CARD_SECURITY_ERROR;  /* Wrong CRC */

  /* Actual length is... */
  *real_length = recv_length;
  return CARDAPPZ_OK;
}

int DesfireDecipherData(CARDAPPZ_CTX_ST * ctx, unsigned char *data,
                        int recv_length, int *real_length, int padd80)
{
  assert(ctx != NULL);

  if (ctx->card_context.desfire.auth_with_sam)
  {
    return CARDAPPZ_NOT_YET_IMPLEMENTED;
  } else
  {
    return DesfireDecipherDataEx(&ctx->card_context.desfire.tdes_ctx, data,
                                 recv_length, real_length, padd80);
  }
}

/**f* Desfire/DesfireEncipherDataEx
 *
 * NAME
 *   DesfireEncipherDataEx
 *
 * DESCRIPTION
 *   Desfire Triple-DES send operation
 *   Append CRC and padding, encipher frame
 *
 * SYNOPSIS
 *   void DesfireEncipherDataEx(TDES_CTX_ST   *tdes_ctx,
 *                     [in/out] unsigned char *data,
 *                              int           recv_length,
 *                     [output] int           *total_length);
 *
 **/
void DesfireEncipherDataEx(TDES_CTX_ST * tdes_ctx, unsigned char *data,
                           int data_length, int *total_length)
{
  assert(tdes_ctx != NULL);
  assert(data != NULL);

  /* At first we have to append the 2 CRC bytes */
  SPROX_ComputeIso14443ACrc(&data[data_length], data, (WORD) data_length);
  data_length += 2;

  /* The we add padding to reach a multiple of 8 */
  while (data_length % 8)
    data[data_length++] = 0x00;

  /* Finally we do the decipher operation on the whole buffer */
  TDES_Decrypt_CBC_S(tdes_ctx, data, data_length / 8);

  /* Actual length is... */
  if (total_length != NULL)
    *total_length = data_length;
}

void DesfireEncipherData(CARDAPPZ_CTX_ST * ctx, unsigned char *data,
                         int data_length, int *total_length)
{
  assert(ctx != NULL);

  if (ctx->card_context.desfire.auth_with_sam)
  {

  } else
  {
    DesfireEncipherDataEx(&ctx->card_context.desfire.tdes_ctx, data,
                          data_length, total_length);
  }
}

/**f* Desfire/DesfireGenerateMACEx
 *
 * NAME
 *   DesfireGenerateMACEx
 *
 * DESCRIPTION
 *   Compute the MAC of a frame
 *
 * SYNOPSIS
 *   void DesfireGenerateMAC(TDES_CTX_ST         *tdes_ctx,
 *                   [input] const unsigned char *data,
 *                           int                 data_length
 *                  [output] unsigned char       mac[4]);
 *
 *
 **/
void DesfireGenerateMACEx(TDES_CTX_ST * tdes_ctx, const unsigned char *data,
                          int data_length, unsigned char mac[4])
{
  int i, j;
  int blocks;
  unsigned char buffer[8];
  unsigned char carry[8];

  assert(tdes_ctx != NULL);
  assert(data != NULL);
  assert(mac != NULL);

  blocks = (7 + data_length) / 8;

  memset(carry, 0x00, 8);       /* IV <- 00...00   */
  for (i = 0; i < blocks; i++)
  {
    for (j = 0; j < 8; j++)
    {
      if (8 * i + j >= data_length)
        buffer[j] = 0;
      else
        buffer[j] = data[8 * i + j];
      buffer[j] ^= carry[j];    /* P  <- P XOR IV  */
    }
    TDES_Encrypt(tdes_ctx, buffer); /* P <- i3DES(P)   */
    memcpy(carry, buffer, 8);   /* P  <- IV        */
  }

  memcpy(mac, &buffer[0], 4);
}

void DesfireGenerateMAC(CARDAPPZ_CTX_ST * ctx, const unsigned char *data,
                        int data_length, unsigned char mac[4])
{
  assert(ctx != NULL);

  if (ctx->card_context.desfire.auth_with_sam)
  {
    /* Use the SAM to generate the MAC */
    DesfireSAMGenerateMAC(ctx, data, data_length, mac);
  } else
  {
    DesfireGenerateMACEx(&ctx->card_context.desfire.tdes_ctx, data,
                         data_length, mac);
  }
}

/**f* Desfire/DesfireVerifyMACEx
 *
 * NAME
 *   DesfireVerifyMACEx
 *
 * DESCRIPTION
 *   Check MAC of a received frame
 *
 * SYNOPSIS
 *   int DesfireVerifyMACEx(TDES_CTX_ST         *tdes_ctx,
 *                  [input] const unsigned char *data,
 *                          int                 recv_length);
 *
 **/
int DesfireVerifyMACEx(TDES_CTX_ST * tdes_ctx, const unsigned char *data,
                       int recv_length)
{
  unsigned char mac[4];

  assert(tdes_ctx != NULL);
  assert(data != NULL);

  if (recv_length <= 4)
    return CARDAPPZ_CARD_SECURITY_ERROR;

  DesfireGenerateMACEx(tdes_ctx, data, recv_length - 4, mac);

  if (memcmp(mac, &data[recv_length - 4], 4))
    return CARDAPPZ_CARD_SECURITY_ERROR;

  return CARDAPPZ_OK;
}

int DesfireVerifyMAC(CARDAPPZ_CTX_ST * ctx, const unsigned char *data,
                     int recv_length)
{
  assert(ctx != NULL);

  if (ctx->card_context.desfire.auth_with_sam)
  {
    /* Use the SAM to verify the MAC */
    return DesfireSAMVerifyMAC(ctx, data, recv_length);

  } else
  {
    return DesfireVerifyMACEx(&ctx->card_context.desfire.tdes_ctx, data,
                              recv_length);
  }
}

static int DesfireTransmitEx(CARDAPPZ_CTX_ST * ctx, unsigned char cmd[],
                             int cmdlen, unsigned char rsp[], int *rsplen)
{
  SWORD rc;

  assert(ctx != NULL);

  if (ctx->trace >= DUMPMORE)
  {
    int i;

    printf("APDU<");
    for (i = 0; i < cmdlen; i++)
      printf("%02X", cmd[i]);
    printf("\n");
  }

  /* T=CL transmit using SpringProx reader  */
  /* not difficult to change for PC/SC mode */
  if (rsplen != NULL)
  {
    WORD t = (WORD) * rsplen;

    rc = SPROX_TclA_Exchange(0xFF, 1, 0xFF, cmd, (WORD) cmdlen, rsp, &t);
    *rsplen = t;
  } else
  {
    rc = SPROX_TclA_Exchange(0xFF, 1, 0xFF, cmd, (WORD) cmdlen, rsp, NULL);
  }

  if (ctx->trace >= DUMPMORE)
  {
    int i;

    printf("APDU>");
    if ((rsp != NULL) && (rsplen != NULL))
      for (i = 0; i < *rsplen; i++)
        printf("%02X", rsp[i]);
    printf("\n");
  }

  return rc;
}

/**f* Desfire/DesfireTransmit
 *
 * NAME
 *   DesfireTransmit
 *
 * DESCRIPTION
 *   Send a command to a Desfire card, a receive its answer
 *
 * SYNOPSIS
 *   int DesfireTransmit(CARDAPPZ_CTX_ST *ctx, unsigned char cmd[], int cmdlen, unsigned char rsp[], int *rsplen, int check_answer);
 *
 * INPUTS
 *   CARDAPPZ_CTX_ST *ctx              : pointer to the global context
 *   unsigned char cmd[]               : the Desfire command
 *   int           cmdlen              : size of command
 *   unsigned char rsp[]               : [output] the Desfire answer      (optional)
 *   int           *rsplen             : [input]  size of answer buffer   (optional)
 *   int           *rsplen             : [output] actuel received length  (optional)
 *   int           check_answer        : 0              : no check, answer is returned without
 *                                                        further processing
 *                                       1 (CHECK_OK)   : function checks that returned code
 *                                                        is 0x00 (Desfire Operation OK)
 *                                       2 (CHECK_MORE) : function checks that returned code
 *                                                        is 0xAF (Desfire Additional Frame)
 *
 * RETURNS
 *   CARDAPPZ_OK                       : success
 *   CARDAPPZ_BUFFER_OVERFLOW          : cmdlen is too big for Desfire card
 *   CARDAPPZ_CARD_PROTOCOL_ERROR      : invalid format of received answer
 *   Other error code from lower layers
 *
 * NOTES
 *   Desfire command must be formated according to Philips Desfire Function Specification :
 *   [COMMAND CODE (mandatory, 1 byte)] [COMMAND DATA (optional, 0 to 59 bytes)]
 *   The native Desfire command is wrapped into an ISO/IEC 7816-4 APDU according to functional
 *   specification 3.1, chapter 3.13 . This works only with Desfire release 6 and later.
 *   To use Desfire release 4, you must rewrite this function to use the legacy mode.
 *
 **/
int DesfireTransmit(CARDAPPZ_CTX_ST * ctx, unsigned char cmd[], int cmdlen,
                    unsigned char rsp[], int *rsplen, int check_answer)
{
  unsigned char io_buffer[60 + 5];
  int o_len, i_len;
  unsigned char sw;
  int rc;

  if (cmd == NULL)
    return CARDAPPZ_INVALID_PARAMETER;
  if (cmdlen < 1)
    return CARDAPPZ_INVALID_PARAMETER;

  if (ctx->trace >= DUMPMORE)
  {
    int i;

    printf("DesF<");
    for (i = 0; i < cmdlen; i++)
      printf("%02X", cmd[i]);
    printf("\n");
  }

  /* Wrapping of Desfire command and data into an ISO 7816-4 APDU */
  io_buffer[0] = 0x90;
  io_buffer[1] = cmd[0];
  io_buffer[2] = 0x00;
  io_buffer[3] = 0x00;

  if (cmdlen == 1)
  {
    o_len = 5;
  } else
  {
    o_len = 5 + cmdlen;
    if (o_len >= sizeof(io_buffer))
      return CARDAPPZ_BUFFER_OVERFLOW;
    io_buffer[4] = (BYTE) (cmdlen - 1);
    memcpy(&io_buffer[5], &cmd[1], cmdlen - 1);
  }
  io_buffer[o_len - 1] = 0x00;

  i_len = sizeof(io_buffer);
  if (rsplen != NULL)
    if (i_len > (*rsplen + 1))
      i_len = (*rsplen + 1);

  rc = DesfireTransmitEx(ctx, io_buffer, o_len, io_buffer, &i_len);
  if (rc)
    return rc;

  if (i_len < 2)
    return CARDAPPZ_CARD_PROTOCOL_ERROR;
  if (io_buffer[i_len - 2] != 0x91)
    return CARDAPPZ_CARD_PROTOCOL_ERROR;

  sw = io_buffer[i_len - 1];

  if (ctx->trace >= DUMPMORE)
  {
    int i;

    printf("DesF>");
    printf("%02X", sw);
    for (i = 0; i < i_len - 2; i++)
      printf("%02X", io_buffer[i]);
    printf("\n");
  }

  if (rsp != NULL)
  {
    rsp[0] = sw;
    memcpy(&rsp[1], &io_buffer[0], i_len - 2);
    if (rsplen != NULL)
      *rsplen = i_len - 1;
  }

  if (check_answer && CHECK_OK)
    if (sw == 0x00)
      return CARDAPPZ_OK;
  if (check_answer && CHECK_MORE)
    if (sw == 0xAF)
      return CARDAPPZ_OK;

  if (ctx->trace >= TRACE)
    printf("Desfire: card error %02X\n", sw);

  return (-1000 - sw);
}

/**f* Desfire/DesfireAuthenticate_1
 *
 * NAME
 *   DesfireAuthenticate_1
 *
 * DESCRIPTION
 *   Desfire mutual authentication command, step 1
 *   (functional specification 3.1, chapter 4.3.1)
 *
 * SYNOPSIS
 *   int DesfireAuthenticate_1(CARDAPPZ_CTX_ST *ctx,
 *                             unsigned char   KeyIndex,
 *                    [output] unsigned char   PiccChallenge[8]);
 *
 **/
int DesfireAuthenticate_1(CARDAPPZ_CTX_ST * ctx,
                          unsigned char KeyIndex,
                          unsigned char PiccChallenge[8])
{
  unsigned char cmd[2];
  unsigned char rsp[9];
  int rsplen;
  int rc;

  if (PiccChallenge == NULL)
    return CARDAPPZ_INVALID_PARAMETER;

  /* Build command, we want 9 bytes in reply */
  cmd[0] = 0x0A;
  cmd[1] = KeyIndex;
  rsplen = 9;
  rc = DesfireTransmit(ctx, cmd, 2, rsp, &rsplen, CHECK_MORE);
  if (rc)
    return rc;
  if (rsplen != 9)
    return CARDAPPZ_CARD_PROTOCOL_ERROR;

  /* OK, we've received the challenge */
  memcpy(PiccChallenge, &rsp[1], 8);
  return CARDAPPZ_OK;
}

/**f* Desfire/DesfireAuthenticate_2
 *
 * NAME
 *   DesfireAuthenticate_2
 *
 * DESCRIPTION
 *   Desfire mutual authentication command, step 2
 *   (functional specification 3.1, chapter 4.3.1)
 *
 * SYNOPSIS
 *   int DesfireAuthenticate_2(CARDAPPZ_CTX_ST *ctx,
 *                             unsigned char   HostCryptogram[16],
 *                             unsigned char   PiccCryptogram[8]);
 *
 **/
int DesfireAuthenticate_2(CARDAPPZ_CTX_ST * ctx,
                          unsigned char HostCryptogram[16],
                          unsigned char PiccCryptogram[8])
{
  unsigned char cmd[17];
  unsigned char rsp[9];
  int rsplen;
  int rc;

  if (HostCryptogram == NULL)
    return CARDAPPZ_INVALID_PARAMETER;
  if (PiccCryptogram == NULL)
    return CARDAPPZ_INVALID_PARAMETER;

  /* Build command, we want 9 bytes in reply */
  cmd[0] = 0xAF;
  memcpy(&cmd[1], HostCryptogram, 16);
  rc = DesfireTransmit(ctx, cmd, 17, rsp, &rsplen, CHECK_OK);
  if (rc)
    return rc;
  if (rsplen != 9)
    return CARDAPPZ_CARD_PROTOCOL_ERROR;

  /* OK, we've received the cryptogram from the card */
  memcpy(PiccCryptogram, &rsp[1], 8);
  return CARDAPPZ_OK;
}

/**f* Desfire/DesfireAuthenticate
 *
 * NAME
 *   DesfireAuthenticate
 *
 * DESCRIPTION
 *   Desfire mutual authentication command (CLA=90, INS=0A)
 *   (functional specification 3.1, chapter 4.3.1)
 *   This calls DesfireAuthenticate_1 followed by DesfireAuthenticate_2
 *   If result is OK,
 *   - the Desfire card can be trusted (it actually knows the key)
 *   - the Desfire card trusts the host application
 *   - a shared session key has been established between the SAM and the
 *     host application to protect (MAC or cipher) further communications
 *
 * SYNOPSIS
 *   int DesfireAuthenticate(CARDAPPZ_CTX_ST     *ctx,
 *                           unsigned char       KeyIndex,
 *                   [input] const unsigned char KeyValue[16]);
 *
 **/
int DesfireAuthenticate(CARDAPPZ_CTX_ST * ctx, unsigned char KeyIndex,
                        const unsigned char KeyValue[16])
{
  int is_tdes;
  int rc;
  int i;
  unsigned char RndA[8], RndB[8];
  unsigned char buffer[16];
  TDES_CTX_ST tdes_ctx;

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireAuthenticate with key %d\n", KeyIndex);

  if (KeyValue == NULL)
    return CARDAPPZ_INVALID_PARAMETER;

  /* Loose current authentication status */
  ctx->card_context.desfire.auth_status = 0;
  ctx->card_context.desfire.auth_with_sam = 0;
  ctx->card_context.desfire.auth_key = KeyIndex;

  /* 3-DES or single DES ? */
  is_tdes = memcmp(&KeyValue[0], &KeyValue[8], 8);

  /* Preload 3-DES context with key */
  TDES_Init(&tdes_ctx, &KeyValue[0], &KeyValue[8], &KeyValue[0]);

  /* Get 8 random bytes as RndA */
  for (i = 0; i < 8; i++)
    RndA[i] = rand();
  /* Note : C library rand() is a WEAK pseudo-random number */
  /* generator, use a strong PRNG instead when available !  */

  /* First step : ask card to perform authentication, */
  /* retrieve its challenge which is Ek(RndB)         */
  rc = DesfireAuthenticate_1(ctx, KeyIndex, RndB);
  if (rc)
    return rc;
  TDES_Decrypt(&tdes_ctx, RndB);

  /* Create our cryptogram : Ek(RndA || RndB') */
  memcpy(&buffer[0], RndA, 8);
  memcpy(&buffer[8], &RndB[1], 7);
  buffer[15] = RndB[0];
  TDES_Decrypt_CBC_S(&tdes_ctx, buffer, 2);

  /* Second step */
  rc = DesfireAuthenticate_2(ctx, buffer, buffer);
  if (rc)
    return rc;

  /* Retrieve RndA' in answer */
  TDES_Decrypt(&tdes_ctx, buffer);
  if (memcmp(&buffer[0], &RndA[1], 7) || (buffer[7] != RndA[0]))
    return CARDAPPZ_CARD_AUTH_FAILURE;  /* Card has not the appropriate key !!! */

  if (is_tdes)
  {
    /* Build the 3-DES session key */
    memcpy(&buffer[0], &RndA[0], 4);
    memcpy(&buffer[4], &RndB[0], 4);
    memcpy(&buffer[8], &RndA[4], 4);
    memcpy(&buffer[12], &RndB[4], 4);
  } else
  {
    /* Build the single DES session key */
    memcpy(&buffer[0], &RndA[0], 4);
    memcpy(&buffer[4], &RndB[0], 4);
    memcpy(&buffer[8], &RndA[0], 4);
    memcpy(&buffer[12], &RndB[0], 4);
  }

  TDES_Init(&ctx->card_context.desfire.tdes_ctx, &buffer[0], &buffer[8],
            &buffer[0]);
  ctx->card_context.desfire.auth_status = 1;

  /* OK */
  return CARDAPPZ_OK;
}

/**f* Desfire/DesfireAuthenticateWithSAM
 *
 * NAME
 *   DesfireAuthenticateWithSAM
 *
 * DESCRIPTION
 *   Desfire mutual authentication command using a Desfire SAM
 *   This calls
 *   - DesfireAuthenticate_1
 *   - DesfireSAMAuthenticatePICC_1
 *   - DesfireAuthenticate_2
 *   - DesfireSAMAuthenticatePICC_2
 *   
 *   If result is OK,
 *   - the Desfire card can be trusted (it actually knows the key)
 *   - the Desfire card trusts the host application (thanks to the SAM)
 *   - a shared session key has been established between the SAM and the
 *     card (but unknown to the host application) to protect (MAC or cipher)
 *     further communications
 *
 * SYNOPSIS
 *   int DesfireAuthenticateWithSam(CARDAPPZ_CTX_ST *ctx,
 *                                  unsigned char   PiccKeyIndex,
 *                                  unsigned char   SamKeyIndex,
 *                                  unsigned char   SamKeyVersion);
 *
 **/
int DesfireAuthenticateWithSam(CARDAPPZ_CTX_ST * ctx,
                               unsigned char PiccKeyIndex,
                               unsigned char SamKeyIndex,
                               unsigned char SamKeyVersion)
{
  unsigned char PiccChallenge[8];
  unsigned char SamCryptogram[16];
  unsigned char PiccCryptogram[8];
  int rc;

  assert(ctx != NULL);

  if (ctx->trace >= TRACE)
    printf("DesfireAuthenticate with key %d (%d / %02X in SAM)\n",
           PiccKeyIndex, SamKeyIndex, SamKeyVersion);

  /* Loose current authentication status */
  ctx->card_context.desfire.auth_status = 0;
  ctx->card_context.desfire.auth_with_sam = 1;
  ctx->card_context.desfire.auth_key = PiccKeyIndex;

  rc = DesfireAuthenticate_1(ctx, PiccKeyIndex, PiccChallenge);
  if (rc)
    return rc;

  rc =
    DesfireSAMAuthenticatePICC_1(ctx, SamKeyIndex, SamKeyVersion,
                                 PiccChallenge, SamCryptogram);
  if (rc)
    return rc;

  rc = DesfireAuthenticate_2(ctx, SamCryptogram, PiccCryptogram);
  if (rc)
    return rc;

  rc = DesfireSAMAuthenticatePICC_2(ctx, PiccCryptogram);
  if (rc)
    return rc;

  /* OK */
  ctx->card_context.desfire.auth_status = 1;
  return CARDAPPZ_OK;
}

/**f* Desfire/DesfireChangeKey
 *
 * NAME
 *   DesfireChangeKey
 *
 * DESCRIPTION
 *   Desfire Change Key command
 *   (functional specification 3.1, chapter 4.3.4)
 *
 * SYNOPSIS
 *   int DesfireChangeKey(CARDAPPZ_CTX_ST *ctx, unsigned char KeyIndex, const unsigned char KeyValue[16]);
 *
 **/
int DesfireChangeKey(CARDAPPZ_CTX_ST * ctx, unsigned char KeyIndex,
                     const unsigned char KeyValue[16])
{
  unsigned char cmd[26];

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireChangeKey %d\n", KeyIndex);

  if (KeyValue == NULL)
    return CARDAPPZ_INVALID_PARAMETER;

  cmd[0] = 0xC4;
  cmd[1] = KeyIndex;
  memcpy(&cmd[2], KeyValue, 16);
  SPROX_ComputeIso14443ACrc(&cmd[18], KeyValue, 16);
  memset(&cmd[20], 0, 6);
  TDES_Decrypt_CBC_S(&ctx->card_context.desfire.tdes_ctx, &cmd[2], 3);

  return DesfireTransmit(ctx, cmd, 26, NULL, NULL, CHECK_OK);
}

/**f* Desfire/DesfireCreateApplication
 *
 * NAME
 *   DesfireCreateApplication
 *
 * DESCRIPTION
 *   Desfire Create Application (= create DF) command
 *   (functional specification 3.1, chapter 4.4.1)
 *
 * SYNOPSIS
 *   int DesfireCreateApplication(CARDAPPZ_CTX_ST *ctx, unsigned long Aid, unsigned char KeySettings, unsigned char KeyCount);
 *
 **/
int DesfireCreateApplication(CARDAPPZ_CTX_ST * ctx, unsigned long Aid,
                             unsigned char KeySettings,
                             unsigned char KeyCount)
{
  unsigned char cmd[6];

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireCreateApplication %06lX\n", Aid);

  cmd[0] = 0xCA;
  /* AID is transmitted LSB first */
  cmd[1] = (unsigned char) (Aid % 0x00000100);
  Aid /= 0x00000100;
  cmd[2] = (unsigned char) (Aid % 0x00000100);
  Aid /= 0x00000100;
  cmd[3] = (unsigned char) (Aid % 0x00000100);
  cmd[4] = KeySettings;
  cmd[5] = KeyCount;

  return DesfireTransmit(ctx, cmd, 6, NULL, NULL, CHECK_OK);
}

/**f* Desfire/DesfireSelectApplication
 *
 * NAME
 *   DesfireSelectApplication
 *
 * DESCRIPTION
 *   Desfire Select Application (= select DF under MF) command
 *   (functional specification 3.1, chapter 4.4.4)
 *
 * SYNOPSIS
 *   int DesfireSelectApplication(CARDAPPZ_CTX_ST *ctx, unsigned long Aid);
 *
 **/

int DesfireSelectApplication(CARDAPPZ_CTX_ST * ctx, unsigned long Aid)
{
  unsigned char cmd[4];

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireSelectApplication %06lX\n", Aid);

  /* Loose current authentication status */
  ctx->card_context.desfire.auth_status = 0;

  cmd[0] = 0x5A;
  /* AID is transmitted LSB first */
  cmd[1] = (unsigned char) (Aid % 0x00000100);
  Aid /= 0x00000100;
  cmd[2] = (unsigned char) (Aid % 0x00000100);
  Aid /= 0x00000100;
  cmd[3] = (unsigned char) (Aid % 0x00000100);

  return DesfireTransmit(ctx, cmd, 4, NULL, NULL, CHECK_OK);
}

/**f* Desfire/DesfireFormatPICC
 *
 * NAME
 *   DesfireFormatPICC
 *
 * DESCRIPTION
 *   Desfire Format PICC command
 *   (functional specification 3.1, chapter 4.4.5)
 *
 * SYNOPSIS
 *   int DesfireFormatPICC(CARDAPPZ_CTX_ST *ctx);
 *
 **/
int DesfireFormatPICC(CARDAPPZ_CTX_ST * ctx)
{
  BYTE cmd[] = { 0xFC };

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireFormatPICC\n");

  return DesfireTransmit(ctx, cmd, sizeof(cmd), NULL, NULL, CHECK_OK);
}

/**f* Desfire/DesfireCreateStdDataFile
 *
 * NAME
 *   DesfireCreateStdDataFile
 *
 * DESCRIPTION
 *   Desfire Create Standard Data File (= transparent EF) command
 *   (functional specification 3.1, chapter 4.5.4)
 *
 * SYNOPSIS
 *   int DesfireCreateStdDataFile(CARDAPPZ_CTX_ST *ctx, unsigned char FileIndex, unsigned char CommSettings, unsigned short AccessRights, unsigned long FileSize);
 *
 **/
int DesfireCreateStdDataFile(CARDAPPZ_CTX_ST * ctx, unsigned char FileIndex,
                             unsigned char CommSettings,
                             unsigned short AccessRights,
                             unsigned long FileSize)
{
  unsigned char cmd[8];

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireCreateStdDataFile %02lX\n", FileIndex);

  cmd[0] = 0xCD;
  cmd[1] = FileIndex;
  cmd[2] = CommSettings;
  cmd[3] = (unsigned char) (AccessRights % 0x0100);
  AccessRights /= 0x0100;
  cmd[4] = (unsigned char) (AccessRights % 0x0100);
  cmd[5] = (unsigned char) (FileSize % 0x00000100);
  FileSize /= 0x00000100;
  cmd[6] = (unsigned char) (FileSize % 0x00000100);
  FileSize /= 0x00000100;
  cmd[7] = (unsigned char) (FileSize % 0x00000100);

  return DesfireTransmit(ctx, cmd, 8, NULL, NULL, CHECK_OK);
}

/**f* Desfire/DesfireCreateValueFile
 *
 * NAME
 *   DesfireCreateValueFile
 *
 * DESCRIPTION
 *   Desfire Create Value File (= counter EF) command
 *   (functional specification 3.1, chapter 4.5.6)
 *
 * SYNOPSIS
 *   int DesfireCreateValueFile(CARDAPPZ_CTX_ST *ctx, unsigned char FileIndex, unsigned char CommSettings, unsigned short AccessRights, signed long LowerLimit, signed long UpperLimit, signed long CurrentValue, unsigned char LimitedCredit);
 *
 **/
int DesfireCreateValueFile(CARDAPPZ_CTX_ST * ctx, unsigned char FileIndex,
                           unsigned char CommSettings,
                           unsigned short AccessRights,
                           signed long LowerLimit, signed long UpperLimit,
                           signed long CurrentValue,
                           unsigned char LimitedCredit)
{
  unsigned char cmd[18];

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireCreateValueFile %02lX\n", FileIndex);

  cmd[0] = 0xCC;
  cmd[1] = FileIndex;
  cmd[2] = CommSettings;
  cmd[3] = (unsigned char) (AccessRights % 0x0100);
  AccessRights /= 0x0100;
  cmd[4] = (unsigned char) (AccessRights % 0x0100);
  cmd[5] = (unsigned char) (LowerLimit % 0x00000100);
  LowerLimit /= 0x00000100;
  cmd[6] = (unsigned char) (LowerLimit % 0x00000100);
  LowerLimit /= 0x00000100;
  cmd[7] = (unsigned char) (LowerLimit % 0x00000100);
  LowerLimit /= 0x00000100;
  cmd[8] = (unsigned char) (LowerLimit % 0x00000100);
  cmd[9] = (unsigned char) (UpperLimit % 0x00000100);
  UpperLimit /= 0x00000100;
  cmd[10] = (unsigned char) (UpperLimit % 0x00000100);
  UpperLimit /= 0x00000100;
  cmd[11] = (unsigned char) (UpperLimit % 0x00000100);
  UpperLimit /= 0x00000100;
  cmd[12] = (unsigned char) (UpperLimit % 0x00000100);
  cmd[13] = (unsigned char) (CurrentValue % 0x00000100);
  CurrentValue /= 0x00000100;
  cmd[14] = (unsigned char) (CurrentValue % 0x00000100);
  CurrentValue /= 0x00000100;
  cmd[15] = (unsigned char) (CurrentValue % 0x00000100);
  CurrentValue /= 0x00000100;
  cmd[16] = (unsigned char) (CurrentValue % 0x00000100);
  cmd[17] = LimitedCredit;

  return DesfireTransmit(ctx, cmd, 18, NULL, NULL, CHECK_OK);
}

/**f* Desfire/DesfireCreateLinearRecordFile
 *
 * NAME
 *   DesfireCreateLinearRecordFile
 *
 * DESCRIPTION
 *   Desfire Create Linear Record File (= linear EF) command
 *   (functional specification 3.1, chapter 4.5.7)
 *
 * SYNOPSIS
 *   int DesfireCreateLinearRecordFile(CARDAPPZ_CTX_ST *ctx, unsigned char FileIndex, unsigned char CommSettings, unsigned short AccessRights, unsigned long RecordSize, unsigned long RecordCount);
 *
 **/
int DesfireCreateLinearRecordFile(CARDAPPZ_CTX_ST * ctx,
                                  unsigned char FileIndex,
                                  unsigned char CommSettings,
                                  unsigned short AccessRights,
                                  unsigned long RecordSize,
                                  unsigned long RecordCount)
{
  unsigned char cmd[11];

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireCreateLinearRecordFile %02lX\n", FileIndex);

  cmd[0] = 0xC1;
  cmd[1] = FileIndex;
  cmd[2] = CommSettings;
  cmd[3] = (unsigned char) (AccessRights % 0x0100);
  AccessRights /= 0x0100;
  cmd[4] = (unsigned char) (AccessRights % 0x0100);
  cmd[5] = (unsigned char) (RecordSize % 0x00000100);
  RecordSize /= 0x00000100;
  cmd[6] = (unsigned char) (RecordSize % 0x00000100);
  RecordSize /= 0x00000100;
  cmd[7] = (unsigned char) (RecordSize % 0x00000100);
  cmd[8] = (unsigned char) (RecordCount % 0x00000100);
  RecordCount /= 0x00000100;
  cmd[9] = (unsigned char) (RecordCount % 0x00000100);
  RecordCount /= 0x00000100;
  cmd[10] = (unsigned char) (RecordCount % 0x00000100);

  return DesfireTransmit(ctx, cmd, 11, NULL, NULL, CHECK_OK);
}

/**f* Desfire/DesfireCreateCyclicRecordFile
 *
 * NAME
 *   DesfireCreateCyclicRecordFile
 *
 * DESCRIPTION
 *   Desfire Create Cyclic Record File (= cyclic EF) command
 *   (functional specification 3.1, chapter 4.5.8)
 *
 * SYNOPSIS
 *   int DesfireCreateCyclicRecordFile(CARDAPPZ_CTX_ST *ctx, unsigned char FileIndex, unsigned char CommSettings, unsigned short AccessRights, unsigned long RecordSize, unsigned long RecordCount);
 *
 **/
int DesfireCreateCyclicRecordFile(CARDAPPZ_CTX_ST * ctx,
                                  unsigned char FileIndex,
                                  unsigned char CommSettings,
                                  unsigned short AccessRights,
                                  unsigned long RecordSize,
                                  unsigned long RecordCount)
{
  unsigned char cmd[11];

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireCreateCyclicRecordFile %02lX\n", FileIndex);

  cmd[0] = 0xC0;
  cmd[1] = FileIndex;
  cmd[2] = CommSettings;
  cmd[3] = (unsigned char) (AccessRights % 0x0100);
  AccessRights /= 0x0100;
  cmd[4] = (unsigned char) (AccessRights % 0x0100);
  cmd[5] = (unsigned char) (RecordSize % 0x00000100);
  RecordSize /= 0x00000100;
  cmd[6] = (unsigned char) (RecordSize % 0x00000100);
  RecordSize /= 0x00000100;
  cmd[7] = (unsigned char) (RecordSize % 0x00000100);
  cmd[8] = (unsigned char) (RecordCount % 0x00000100);
  RecordCount /= 0x00000100;
  cmd[9] = (unsigned char) (RecordCount % 0x00000100);
  RecordCount /= 0x00000100;
  cmd[10] = (unsigned char) (RecordCount % 0x00000100);

  return DesfireTransmit(ctx, cmd, 11, NULL, NULL, CHECK_OK);
}

/*
 * Common part for ReadData and ReadRecord
 * ---------------------------------------
 */
int DesfireReadCommon(CARDAPPZ_CTX_ST * ctx, unsigned char CommandCode,
                      unsigned char FileIndex, unsigned char CommMode,
                      unsigned long Offset, unsigned long Length,
                      unsigned char Buffer[], unsigned long *ActualLength)
{
  unsigned char buf[60];
  unsigned char storage[4096];  /* We need a temporary storage to perform MAC or Decipher operations */
  int padd80;
  int rsplen;
  unsigned int gotlen;
  int rc;

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
  {
    if (CommandCode == 0xBD)
    {
      printf("DesfireReadData file=%d, offset=%ld, length=%ld\n", FileIndex,
             Offset, Length);
    } else if (CommandCode == 0xBB)
    {
      printf("DesfireReadRecords file=%d, offset=%ld, length=%ld\n",
             FileIndex, Offset, Length);
    } else
    {
      printf("DesfireReadCommon code=%02X file=%d, offset=%ld, length=%ld\n",
             CommandCode, FileIndex, Offset, Length);
    }
  }

  padd80 = (Length == 0);

  buf[0] = CommandCode;
  buf[1] = FileIndex;
  /* Offset, LSB first */
  buf[2] = (unsigned char) (Offset % 0x00000100);
  Offset /= 0x00000100;
  buf[3] = (unsigned char) (Offset % 0x00000100);
  Offset /= 0x00000100;
  buf[4] = (unsigned char) (Offset % 0x00000100);
  /* Length, LSB first */
  buf[5] = (unsigned char) (Length % 0x00000100);
  Length /= 0x00000100;
  buf[6] = (unsigned char) (Length % 0x00000100);
  Length /= 0x00000100;
  buf[7] = (unsigned char) (Length % 0x00000100);

  rsplen = sizeof(buf);
  rc = DesfireTransmit(ctx, buf, 8, buf, &rsplen, CHECK_OK | CHECK_MORE);
  if (rc)
    return rc;

  memcpy(&storage[0], &buf[1], rsplen - 1);
  gotlen = rsplen - 1;

  while (buf[0] == 0xAF)
  {
    rsplen = sizeof(buf);
    rc = DesfireTransmit(ctx, buf, 1, buf, &rsplen, CHECK_OK | CHECK_MORE);
    if (rc)
      return rc;

    if (Buffer != NULL)
      memcpy(&storage[gotlen], &buf[1], rsplen - 1);

    gotlen += rsplen - 1;
  }

  switch (CommMode)
  {
    case DESFIRE_COMM_PLAIN:
      break;

    case DESFIRE_COMM_MACED:
      rc = DesfireVerifyMAC(ctx, storage, gotlen);
      gotlen -= 4;
      if (rc)
        return rc;
      break;

    case DESFIRE_COMM_CIPHER:
      rc = DesfireDecipherData(ctx, storage, gotlen, &gotlen, padd80);
      if (rc)
        return rc;
      break;

    default:
      return CARDAPPZ_INVALID_PARAMETER;
  }

  if (ActualLength != NULL)
  {
    if (*ActualLength < gotlen)
    {
      if (Buffer != NULL)
        memcpy(Buffer, storage, *ActualLength);
      rc = CARDAPPZ_BUFFER_OVERFLOW;
    } else
    {
      if (Buffer != NULL)
        memcpy(Buffer, storage, gotlen);
    }
    *ActualLength = gotlen;
  } else if (Buffer != NULL)
  {
    memcpy(Buffer, storage, gotlen);
  }

  return rc;
}

/*
 * Common part for WriteData and WriteRecord
 * -----------------------------------------
 */
int DesfireWriteCommon(CARDAPPZ_CTX_ST * ctx, unsigned char CommandCode,
                       unsigned char FileIndex, unsigned char CommMode,
                       unsigned long Offset, unsigned long Length,
                       const unsigned char Buffer[])
{
  unsigned char buf[4096];      /* We need a temporary storage to perform MAC or Encipher operations */
  unsigned int buflen, donelen, nextlen;
  int rc;

  assert(ctx != NULL);
  assert(Buffer != NULL);

  if (ctx->trace >= TRACE)
  {
    if (CommandCode == 0x3D)
    {
      printf("DesfireWriteData file=%d, offset=%ld, length=%ld\n", FileIndex,
             Offset, Length);
    } else if (CommandCode == 0x3B)
    {
      printf("DesfireWriteRecord file=%d, offset=%ld, length=%ld\n",
             FileIndex, Offset, Length);
    } else
    {
      printf
        ("DesfireWriteCommon code=%02X file=%d, offset=%ld, length=%ld\n",
         CommandCode, FileIndex, Offset, Length);
    }
  }

  /* Memorize current length */
  buflen = Length;

  buf[0] = CommandCode;
  buf[1] = FileIndex;
  /* Offset, LSB first */
  buf[2] = (unsigned char) (Offset % 0x00000100);
  Offset /= 0x00000100;
  buf[3] = (unsigned char) (Offset % 0x00000100);
  Offset /= 0x00000100;
  buf[4] = (unsigned char) (Offset % 0x00000100);
  /* Length, LSB first */
  buf[5] = (unsigned char) (Length % 0x00000100);
  Length /= 0x00000100;
  buf[6] = (unsigned char) (Length % 0x00000100);
  Length /= 0x00000100;
  buf[7] = (unsigned char) (Length % 0x00000100);

  /* Data to be written starts at offset 8 */
  memcpy(&buf[8], Buffer, buflen);

  /* Add MAC or cipher the stream */
  switch (CommMode)
  {
    case DESFIRE_COMM_PLAIN:
      break;

    case DESFIRE_COMM_MACED:
      DesfireGenerateMAC(ctx, &buf[8], buflen, &buf[8 + buflen]);
      buflen += 4;
      break;

    case DESFIRE_COMM_CIPHER:
      DesfireEncipherData(ctx, &buf[8], buflen, &buflen);
      break;

    default:
      return CARDAPPZ_INVALID_PARAMETER;
  }

  /* Total size is 8 (for header) + data length */
  buflen += 8;

  /* Send data frame by frame, with a limit of 59 bytes + command code */
  donelen = 0;
  while (donelen < buflen)
  {
    nextlen = buflen - donelen;
    if (nextlen > 59)
    {
      nextlen = 59;
      rc =
        DesfireTransmit(ctx, &buf[donelen], nextlen, NULL, NULL, CHECK_MORE);
    } else
    {
      rc = DesfireTransmit(ctx, &buf[donelen], nextlen, NULL, NULL, CHECK_OK);
    }
    if (rc)
      return rc;

    donelen += nextlen;
    buf[donelen - 1] = 0xAF;
  }

  return CARDAPPZ_OK;
}

/**f* Desfire/DesfireReadData
 *
 * NAME
 *   DesfireReadData
 *
 * DESCRIPTION
 *   Desfire Read Data (= read binary) command
 *   (functional specification 3.1, chapter 4.6.1)
 *
 * SYNOPSIS
 *   int DesfireReadData(CARDAPPZ_CTX_ST *ctx, unsigned char FileIndex, unsigned char CommMode, unsigned long Offset, unsigned long Length, unsigned char Buffer[], unsigned long *ActualLength);
 *
 **/
int DesfireReadData(CARDAPPZ_CTX_ST * ctx, unsigned char FileIndex,
                    unsigned char CommMode, unsigned long Offset,
                    unsigned long Length, unsigned char Buffer[],
                    unsigned long *ActualLength)
{
  return DesfireReadCommon(ctx, 0xBD, FileIndex, CommMode, Offset, Length,
                           Buffer, ActualLength);
}

/**f* Desfire/DesfireWriteData
 *
 * NAME
 *   DesfireWriteData
 *
 * DESCRIPTION
 *   Desfire Write Data (= update binary) command
 *   (functional specification 3.1, chapter 4.6.2)
 *
 * SYNOPSIS
 *   int DesfireWriteData(CARDAPPZ_CTX_ST *ctx, unsigned char FileIndex, unsigned char CommMode, unsigned long Offset, unsigned long Length, const unsigned char Buffer[]);
 *
 **/
int DesfireWriteData(CARDAPPZ_CTX_ST * ctx, unsigned char FileIndex,
                     unsigned char CommMode, unsigned long Offset,
                     unsigned long Length, const unsigned char Buffer[])
{
  return DesfireWriteCommon(ctx, 0x3D, FileIndex, CommMode, Offset, Length,
                            Buffer);
}

/**f* Desfire/DesfireGetValue
 *
 * NAME
 *   DesfireGetValue
 *
 * DESCRIPTION
 *   Desfire Get Value (= read counter) command
 *   (functional specification 3.1, chapter 4.6.3)
 *
 * SYNOPSIS
 *   int DesfireGetValue(CARDAPPZ_CTX_ST *ctx, unsigned char FileIndex, unsigned char CommMode, signed long *Value);
 *
 **/
int DesfireGetValue(CARDAPPZ_CTX_ST * ctx, unsigned char FileIndex,
                    unsigned char CommMode, signed long *Value)
{
  unsigned char buf[9];
  int rsplen = sizeof(buf);
  int rc;
  signed long v;

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireGetValue file=%d\n", FileIndex);

  buf[0] = 0x6C;
  buf[1] = FileIndex;

  rc = DesfireTransmit(ctx, buf, 2, buf, &rsplen, CHECK_OK);
  if (rc)
    return rc;

  switch (CommMode)
  {
    case DESFIRE_COMM_PLAIN:
      if (rsplen != 5)
        return CARDAPPZ_CARD_PROTOCOL_ERROR;
      break;

    case DESFIRE_COMM_MACED:
      if (rsplen != 9)
        return CARDAPPZ_CARD_PROTOCOL_ERROR;
      rc = DesfireVerifyMAC(ctx, &buf[1], 8);
      if (rc)
        return rc;
      break;

    case DESFIRE_COMM_CIPHER:
      if (rsplen != 9)
        return CARDAPPZ_CARD_PROTOCOL_ERROR;
      rc = DesfireDecipherData(ctx, &buf[1], 8, &rsplen, 0);
      if (rc)
        return rc;
      if (rsplen != 4)
        return CARDAPPZ_CARD_PROTOCOL_ERROR;
      break;

    default:
      return CARDAPPZ_INVALID_PARAMETER;
  }

  /* Value is received LSB first */
  v = buf[4];
  v *= 0x00000100;
  v |= buf[3];
  v *= 0x00000100;
  v |= buf[2];
  v *= 0x00000100;
  v |= buf[1];

  if (Value != NULL)
    *Value = v;

  return CARDAPPZ_OK;
}

/**f* Desfire/DesfireCredit
 *
 * NAME
 *   DesfireCredit
 *
 * DESCRIPTION
 *   Desfire Credit (= increase counter) command
 *   (functional specification 3.1, chapter 4.6.4)
 *
 * SYNOPSIS
 *   int DesfireCredit(CARDAPPZ_CTX_ST *ctx, unsigned char FileIndex, unsigned char CommMode, signed long Value);
 *
 **/
int DesfireCredit(CARDAPPZ_CTX_ST * ctx, unsigned char FileIndex,
                  unsigned char CommMode, signed long Value)
{
  BYTE cmd[10];
  BYTE cmdlen;

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireCredit file=%d value=%ld\n", FileIndex, Value);

  cmd[0] = 0x0C;
  cmd[1] = FileIndex;
  /* Value is transmitted LSB first */
  cmd[2] = (unsigned char) (Value % 0x00000100);
  Value /= 0x00000100;
  cmd[3] = (unsigned char) (Value % 0x00000100);
  Value /= 0x00000100;
  cmd[4] = (unsigned char) (Value % 0x00000100);
  Value /= 0x00000100;
  cmd[5] = (unsigned char) (Value % 0x00000100);

  /* Add MAC or cipher the stream */
  switch (CommMode)
  {
    case DESFIRE_COMM_PLAIN:
      cmdlen = 6;
      break;

    case DESFIRE_COMM_MACED:
      DesfireGenerateMAC(ctx, &cmd[2], 4, &cmd[6]);
      cmdlen = 10;
      break;

    case DESFIRE_COMM_CIPHER:
      DesfireEncipherData(ctx, &cmd[2], 4, NULL);
      cmdlen = 10;
      break;

    default:
      return CARDAPPZ_INVALID_PARAMETER;
  }


  return DesfireTransmit(ctx, cmd, sizeof(cmd), NULL, NULL, CHECK_OK);
}

/**f* Desfire/DesfireDebit
 *
 * NAME
 *   DesfireDebit
 *
 * DESCRIPTION
 *   Desfire Debit (= decrease counter) command
 *   (functional specification 3.1, chapter 4.6.5)
 *
 * SYNOPSIS
 *   int DesfireDebit(CARDAPPZ_CTX_ST *ctx, unsigned char FileIndex, unsigned char CommMode, signed long Value);
 *
 **/
int DesfireDebit(CARDAPPZ_CTX_ST * ctx, unsigned char FileIndex,
                 unsigned char CommMode, signed long Value)
{
  BYTE cmd[10];
  BYTE cmdlen;

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireDebit file=%d value=%ld\n", FileIndex, Value);

  cmd[0] = 0xDC;
  cmd[1] = FileIndex;
  /* Value is transmitted LSB first */
  cmd[2] = (unsigned char) (Value % 0x00000100);
  Value /= 0x00000100;
  cmd[3] = (unsigned char) (Value % 0x00000100);
  Value /= 0x00000100;
  cmd[4] = (unsigned char) (Value % 0x00000100);
  Value /= 0x00000100;
  cmd[5] = (unsigned char) (Value % 0x00000100);

  /* Add MAC or cipher the stream */
  switch (CommMode)
  {
    case DESFIRE_COMM_PLAIN:
      cmdlen = 6;
      break;

    case DESFIRE_COMM_MACED:
      DesfireGenerateMAC(ctx, &cmd[2], 4, &cmd[6]);
      cmdlen = 10;
      break;

    case DESFIRE_COMM_CIPHER:
      DesfireEncipherData(ctx, &cmd[2], 4, NULL);
      cmdlen = 10;
      break;

    default:
      return CARDAPPZ_INVALID_PARAMETER;
  }


  return DesfireTransmit(ctx, cmd, sizeof(cmd), NULL, NULL, CHECK_OK);
}

/**f* Desfire/DesfireWriteRecord
 *
 * NAME
 *   DesfireWriteRecord
 *
 * DESCRIPTION
 *   Desfire Write Record (= append record) command
 *   (functional specification 3.1, chapter 4.6.7)
 *
 * SYNOPSIS
 *   int DesfireWriteRecord(CARDAPPZ_CTX_ST *ctx, unsigned char FileIndex, unsigned char CommMode, unsigned long Offset, unsigned long Length, const unsigned char Buffer[]);
 *
 **/
int DesfireWriteRecord(CARDAPPZ_CTX_ST * ctx, unsigned char FileIndex,
                       unsigned char CommMode, unsigned long Offset,
                       unsigned long Length, const unsigned char Buffer[])
{
  return DesfireWriteCommon(ctx, 0x3B, FileIndex, CommMode, Offset, Length,
                            Buffer);
}

/**f* Desfire/DesfireReadRecords
 *
 * NAME
 *   DesfireReadRecords
 *
 * DESCRIPTION
 *   Desfire Read Records (= read record) command
 *   (functional specification 3.1, chapter 4.6.8)
 *
 * SYNOPSIS
 *   int DesfireReadRecords(CARDAPPZ_CTX_ST *ctx, unsigned char FileIndex, unsigned char CommMode, unsigned long Offset, unsigned long Length, unsigned char Buffer[], unsigned long *ActualLength);
 *
 **/
int DesfireReadRecords(CARDAPPZ_CTX_ST * ctx, unsigned char FileIndex,
                       unsigned char CommMode, unsigned long Offset,
                       unsigned long Length, unsigned char Buffer[],
                       unsigned long *ActualLength)
{
  return DesfireReadCommon(ctx, 0xBB, FileIndex, CommMode, Offset, Length,
                           Buffer, ActualLength);
}

/**f* Desfire/DesfireAbortTransaction
 *
 * NAME
 *   DesfireAbortTransaction
 *
 * DESCRIPTION
 *   Desfire Abort Transaction (= rollback) command
 *   (functional specification 3.1, chapter 4.6.11)
 *
 * SYNOPSIS
 *   int DesfireAbortTransaction(CARDAPPZ_CTX_ST *ctx);
 *
 **/
int DesfireClearRecordFile(CARDAPPZ_CTX_ST * ctx, unsigned char FileIndex)
{
  BYTE cmd[2];

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireClearRecordFile file=%d\n", FileIndex);

  cmd[0] = 0xEB;
  cmd[1] = FileIndex;

  return DesfireTransmit(ctx, cmd, sizeof(cmd), NULL, NULL, CHECK_OK);
}

/**f* Desfire/DesfireCommitTransaction
 *
 * NAME
 *   DesfireCommitTransaction
 *
 * DESCRIPTION
 *   Desfire Commit Transaction (= commit) command
 *   (functional specification 3.1, chapter 4.6.10)
 *
 * SYNOPSIS
 *   int DesfireCommitTransaction(CARDAPPZ_CTX_ST *ctx);
 *
 **/
int DesfireCommitTransaction(CARDAPPZ_CTX_ST * ctx)
{
  BYTE cmd[] = { 0xC7 };

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireCommitTransaction\n");

  return DesfireTransmit(ctx, cmd, sizeof(cmd), NULL, NULL, CHECK_OK);
}

/**f* Desfire/DesfireAbortTransaction
 *
 * NAME
 *   DesfireAbortTransaction
 *
 * DESCRIPTION
 *   Desfire Abort Transaction (= rollback) command
 *   (functional specification 3.1, chapter 4.6.11)
 *
 * SYNOPSIS
 *   int DesfireAbortTransaction(CARDAPPZ_CTX_ST *ctx);
 *
 **/
int DesfireAbortTransaction(CARDAPPZ_CTX_ST * ctx)
{
  BYTE cmd[] = { 0xA7 };

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("DesfireAbortTransaction\n");

  return DesfireTransmit(ctx, cmd, sizeof(cmd), NULL, NULL, CHECK_OK);
}

/**f* Desfire/DesfireSelectCard
 *
 * NAME
 *   DesfireSelectCard
 *
 * DESCRIPTION
 *   This function sends the "ISO Select Application" command to the card, with
 *   Desfire AID as application identifier (D2 76 00 00 85 01 00).
 *   This ensure that the card is actually a Desfire (v6 or later)
 *   (functional specification 3.1, chapter 4.7.1)
 *
 * SYNOPSIS
 *   int DesfireSelectCard(CARDAPPZ_CTX_ST *ctx);
 *
 **/
int DesfireSelectCard(CARDAPPZ_CTX_ST * ctx)
{
  /* Select Desfire ISO applet */
  BYTE cmd[] =
    { 0x00, 0xA4, 0x04, 0x00, 0x07, 0xD2, 0x76, 0x00, 0x00, 0x85, 0x01,
    0x00
  };
  BYTE rsp[2];
  int rsplen = sizeof(rsp);
  int rc;

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
    printf("Desfire: select ISO 7816-4 applet\n");

  rc = DesfireTransmitEx(ctx, cmd, sizeof(cmd), rsp, &rsplen);
  if (rc)
    return rc;

  if ((rsplen != 2) || (rsp[0] != 0x90) || (rsp[1] != 0x00))
    return CARDAPPZ_WRONG_STATUS_WORD;

  return DesfireSelectApplication(ctx, 0);
}

/*
 * Blank the card
 * --------------
 */
int DesfireBlankCard(CARDAPPZ_CTX_ST * ctx)
{
  int rc;

  /* Select application 0 */
  rc = DesfireSelectApplication(ctx, 0);
  if (rc)
    return rc;

  /* Get authenticated with master key */
  rc = DesfireAuthenticate(ctx, 0, DESFIRE_BLANK_KEY);
  if (rc)
    return rc;

  /* Format PICC */
  rc = DesfireFormatPICC(ctx);
  if (rc)
    return rc;

  return CARDAPPZ_OK;
}
