#include "cardappz.h"
#include "cardappz_desfire.h"

#include <assert.h>

int DesfireSelectTransport(CARDAPPZ_CTX_ST * ctx)
{
  int rc;
  unsigned char key[16];

  assert(ctx != NULL);

  /* Select the application directory */
  rc = DesfireSelectApplication(ctx, TRANSPORT_DESFIRE_AID);
  if (rc)
    return rc;

  if (ctx->desfire_sam.found)
  {
    /* Use Desfire SAM for authentication with key 1 */
    rc = DesfireAuthenticateWithSam(ctx,
                                    1,
                                    TRANSPORT_MASTER_KEY1_IDX,
                                    TRANSPORT_MASTER_KEY1_VER);
    if (rc)
      return rc;

  } else
  {
    /* Compute the diversified key 1 */
    rc = ComputeDivKey(ctx, key, TRANSPORT_MASTER_KEY1);
    if (rc)
      return rc;

    /* Get authenticated with key 1 */
    rc = DesfireAuthenticate(ctx, 1, key);
    if (rc)
      return rc;
  }

  return CARDAPPZ_OK;
}

/* Read different files */
int DesfireReadTransportContract(CARDAPPZ_CTX_ST * ctx, int Index,
                                 TRANSPORT_CONTRACT_ST * Contract)
{
  return DesfireReadRecords(ctx, TRANSPORT_DESFIRE_FID_CONTRACTS,
                            DESFIRE_COMM_MACED, Index, 1,
                            (unsigned char *) Contract, NULL);
}

int DesfireReadTransportCounter(CARDAPPZ_CTX_ST * ctx, int Counter,
                                unsigned long *Value)
{
  int rc;
  signed long v;

  switch (Counter)
  {
    case 1:
      rc =
        DesfireGetValue(ctx, TRANSPORT_DESFIRE_FID_COUNTER1,
                        DESFIRE_COMM_MACED, &v);
      break;
    case 2:
      rc =
        DesfireGetValue(ctx, TRANSPORT_DESFIRE_FID_COUNTER2,
                        DESFIRE_COMM_MACED, &v);
      break;
    case 3:
      rc =
        DesfireGetValue(ctx, TRANSPORT_DESFIRE_FID_COUNTER3,
                        DESFIRE_COMM_MACED, &v);
      break;
    default:
      rc = CARDAPPZ_INVALID_PARAMETER;
  }

  if (rc)
    return rc;

  if (Value != NULL)
    *Value = v;
  return CARDAPPZ_OK;
}

int DesfireReadTransportEvent(CARDAPPZ_CTX_ST * ctx, int Index,
                              TRANSPORT_EVENT_ST * Event)
{
  return DesfireReadRecords(ctx, TRANSPORT_DESFIRE_FID_EVENTS,
                            DESFIRE_COMM_MACED, Index, 1,
                            (unsigned char *) Event, NULL);
}

/* New event + decrement counter if needed */
int DesfireAppendTransportEvent(CARDAPPZ_CTX_ST * ctx,
                                TRANSPORT_EVENT_ST * Event)
{
  int rc;

  if (Event == NULL)
    return CARDAPPZ_INVALID_PARAMETER;

  /* Decrement contract counter if needed */
  if (Event->Charged)
  {
    switch (Event->Counter)
    {
      case 1:
        rc =
          DesfireDebit(ctx, TRANSPORT_DESFIRE_FID_COUNTER1,
                       DESFIRE_COMM_MACED, Event->Charged);
        break;
      case 2:
        rc =
          DesfireDebit(ctx, TRANSPORT_DESFIRE_FID_COUNTER2,
                       DESFIRE_COMM_MACED, Event->Charged);
        break;
      case 3:
        rc =
          DesfireDebit(ctx, TRANSPORT_DESFIRE_FID_COUNTER3,
                       DESFIRE_COMM_MACED, Event->Charged);
        break;
      default:
        rc = CARDAPPZ_OK;
    }
    if (rc)
      goto rollback;
  }

  /* Insert new record */
  rc =
    DesfireWriteRecord(ctx, TRANSPORT_DESFIRE_FID_EVENTS, DESFIRE_COMM_MACED,
                       0, sizeof(TRANSPORT_EVENT_ST),
                       (unsigned char *) Event);
  if (rc)
    goto rollback;

  /* Commit transaction */
  rc = DesfireCommitTransaction(ctx);
  if (rc)
    return rc;

  return CARDAPPZ_OK;

rollback:
  DesfireAbortTransaction(ctx);
  return rc;
}

/* Contract writing */

/* ---------------- */

int DesfireAppendTransportContract(CARDAPPZ_CTX_ST * ctx,
                                   TRANSPORT_CONTRACT_ST * Contract,
                                   unsigned long CounterValue)
{
  unsigned long CurrentValue;
  int rc;
  unsigned char key[16];

  assert(Contract != NULL);

  /* Select the application directory */
  rc = DesfireSelectApplication(ctx, TRANSPORT_DESFIRE_AID);
  if (rc)
    return rc;

  if (ctx->desfire_sam.found)
  {
    /* Use Desfire SAM for authentication with key 0 */
    rc = DesfireAuthenticateWithSam(ctx,
                                    0,
                                    TRANSPORT_MASTER_KEY0_IDX,
                                    TRANSPORT_MASTER_KEY0_VER);
    if (rc)
      return rc;

  } else
  {
    /* Compute the diversified key 0 */
    rc = ComputeDivKey(ctx, key, TRANSPORT_MASTER_KEY0);
    if (rc)
      return rc;

    /* Get authenticated with key 0 */
    rc = DesfireAuthenticate(ctx, 0, key);
    if (rc)
      return rc;
  }

  /* Write new record */
  rc =
    DesfireWriteRecord(ctx, TRANSPORT_DESFIRE_FID_CONTRACTS,
                       DESFIRE_COMM_MACED, 0, sizeof(TRANSPORT_CONTRACT_ST),
                       (unsigned char *) Contract);
  if (rc)
    return rc;

  if ((Contract->Counter >= 1) && (Contract->Counter <= 3))
  {
    /* Read current value of counter */
    rc = DesfireReadTransportCounter(ctx, Contract->Counter, &CurrentValue);
    if (rc)
      goto rollback;

    /* Set new value for counter */
    if (CurrentValue > CounterValue)
    {
      /* Credit counter */
      switch (Contract->Counter)
      {
        case 1:
          rc =
            DesfireCredit(ctx, TRANSPORT_DESFIRE_FID_COUNTER1,
                          DESFIRE_COMM_MACED, CurrentValue - CounterValue);
          break;
        case 2:
          rc =
            DesfireCredit(ctx, TRANSPORT_DESFIRE_FID_COUNTER2,
                          DESFIRE_COMM_MACED, CurrentValue - CounterValue);
          break;
        case 3:
          rc =
            DesfireCredit(ctx, TRANSPORT_DESFIRE_FID_COUNTER3,
                          DESFIRE_COMM_MACED, CurrentValue - CounterValue);
          break;
        default:
          rc = CARDAPPZ_OK;
      }
      if (rc)
        goto rollback;

    } else if (CurrentValue < CounterValue)
    {
      /* Debit counter */
      switch (Contract->Counter)
      {
        case 1:
          rc =
            DesfireDebit(ctx, TRANSPORT_DESFIRE_FID_COUNTER1,
                         DESFIRE_COMM_MACED, CounterValue - CurrentValue);
          break;
        case 2:
          rc =
            DesfireDebit(ctx, TRANSPORT_DESFIRE_FID_COUNTER2,
                         DESFIRE_COMM_MACED, CounterValue - CurrentValue);
          break;
        case 3:
          rc =
            DesfireDebit(ctx, TRANSPORT_DESFIRE_FID_COUNTER3,
                         DESFIRE_COMM_MACED, CounterValue - CurrentValue);
          break;
        default:
          rc = CARDAPPZ_OK;
      }
      if (rc)
        goto rollback;
    }
  }

  /* Commit write record + debit or credit transaction */
  rc = DesfireCommitTransaction(ctx);
  if (rc)
    return rc;

  return CARDAPPZ_OK;

rollback:
  DesfireAbortTransaction(ctx);
  return rc;
}

int DesfireEraseTransportContracts(CARDAPPZ_CTX_ST * ctx)
{
  int rc;
  unsigned char key[16];

  /* Select the application directory */
  rc = DesfireSelectApplication(ctx, TRANSPORT_DESFIRE_AID);
  if (rc)
    return rc;

  /* Compute the diversified key 0 */
  rc = ComputeDivKey(ctx, key, TRANSPORT_MASTER_KEY0);
  if (rc)
    return rc;

  /* Get authenticated with key 0 */
  rc = DesfireAuthenticate(ctx, 0, key);
  if (rc)
    return rc;

  /* Clear the contract file */
  rc = DesfireClearRecordFile(ctx, TRANSPORT_DESFIRE_FID_CONTRACTS);
  if (rc)
    return rc;

  /* Commit */
  rc = DesfireCommitTransaction(ctx);
  if (rc)
    return rc;

  return CARDAPPZ_OK;
}

/* Formatting */

/* ---------- */

int DesfireFormatTransport(CARDAPPZ_CTX_ST * ctx)
{
  int rc;
  unsigned char key[16];

  /* Select application 0 */
  rc = DesfireSelectApplication(ctx, 0);
  if (rc)
    return rc;

  /* Get authenticated with master key */
  rc = DesfireAuthenticate(ctx, 0, DESFIRE_BLANK_KEY);
  if (rc)
    return rc;

  /* Create the transport application with 2 keys */
  rc = DesfireCreateApplication(ctx, TRANSPORT_DESFIRE_AID, 0xE3, 2);
  if (rc)
    return rc;

  /* Select the application */
  rc = DesfireSelectApplication(ctx, TRANSPORT_DESFIRE_AID);
  if (rc)
    return rc;

  /* Formatting is performed with (blank) key 0 */
  rc = DesfireAuthenticate(ctx, 0, DESFIRE_BLANK_KEY);
  if (rc)
    return rc;

  /* Create contracts file (3 records, 32 bytes each) */
  rc = DesfireCreateLinearRecordFile(ctx, TRANSPORT_DESFIRE_FID_CONTRACTS, DESFIRE_COMM_MACED, 0x1F0F, 32, 3);  /* MACed ; Key 0 = R/W ; Key 1 = R */
  if (rc)
    return rc;

  /* Create events file (6 + 1 records, 16 bytes each) */
  rc = DesfireCreateCyclicRecordFile(ctx, TRANSPORT_DESFIRE_FID_EVENTS, DESFIRE_COMM_MACED, 0x110F, 16, 7); /* MACed ; Key 0 = R/W ; Key 1 = R+W */
  if (rc)
    return rc;

  /* Create counter 1, min=0, max=10000, initial value = 0, limited credit disabled */
  rc = DesfireCreateValueFile(ctx, TRANSPORT_DESFIRE_FID_COUNTER1, DESFIRE_COMM_MACED, 0x1F0F, 0, 10000, 0, 0); /* MACed ; Key 0 = R/W ; Key 1 = R */
  if (rc)
    return rc;

  /* Create counter 2, min=0, max=10000, initial value = 0, limited credit disabled */
  rc = DesfireCreateValueFile(ctx, TRANSPORT_DESFIRE_FID_COUNTER2, DESFIRE_COMM_MACED, 0x1F0F, 0, 10000, 0, 0); /* MACed ; Key 0 = R/W ; Key 1 = R */
  if (rc)
    return rc;

  /* Create counter 3, min=0, max=10000, initial value = 0, limited credit disabled */
  rc = DesfireCreateValueFile(ctx, TRANSPORT_DESFIRE_FID_COUNTER3, DESFIRE_COMM_MACED, 0x1F0F, 0, 10000, 0, 0); /* MACed ; Key 0 = R/W ; Key 1 = R */
  if (rc)
    return rc;

  /* Compute the diversified key 0 */
  rc = ComputeDivKey(ctx, key, TRANSPORT_MASTER_KEY0);
  if (rc)
    return rc;

  /* Put the new key 0 */
  rc = DesfireChangeKey(ctx, 0, key);
  if (rc)
    return rc;

  /* Get authenticated with (blank) key 1 */
  rc = DesfireAuthenticate(ctx, 1, DESFIRE_BLANK_KEY);
  if (rc)
    return rc;

  /* Compute the diversified key 1 */
  rc = ComputeDivKey(ctx, key, TRANSPORT_MASTER_KEY1);
  if (rc)
    return rc;

  /* Put the new key 1 */
  rc = DesfireChangeKey(ctx, 1, key);
  if (rc)
    return rc;

  /* Done ! */
  return CARDAPPZ_OK;
}
